""" simpleTALES Implementation

		Copyright (c) 2009 Colin Stewart (http://www.owlfish.com/)
		All rights reserved.

		Redistribution and use in source and binary forms, with or without
		modification, are permitted provided that the following conditions
		are met:
		1. Redistributions of source code must retain the above copyright
		   notice, this list of conditions and the following disclaimer.
		2. Redistributions in binary form must reproduce the above copyright
		   notice, this list of conditions and the following disclaimer in the
		   documentation and/or other materials provided with the distribution.
		3. The name of the author may not be used to endorse or promote products
		   derived from this software without specific prior written permission.

		THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
		IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
		OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
		IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
		INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
		NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
		DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
		THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
		(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
		THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

		If you make any bug fixes or feature enhancements please let me know!

		The classes in this module implement the TALES specification, used
		by the simpleTAL module.

		Module Dependencies: logging
"""

import types, sys

import logging

from .. import simpletal


__version__ = simpletal.__version__

DEFAULTVALUE = "This represents a Default value."

class PathNotFoundException (Exception):
	pass

class ContextContentException (Exception):
	""" This is raised when invalid content has been placed into the Context object.
		For example using non-ascii characters instead of Unicode strings.
	"""
	pass

PATHNOTFOUNDEXCEPTION = PathNotFoundException()

class ContextVariable (BaseException):
	def __init__ (self, value = None):
		self.ourValue = value

	def value (self, currentPath=None):
		if (hasattr (self.ourValue, "__call__")):
			return self.ourValue()
		return self.ourValue

	def rawValue (self):
		return self.ourValue

	def __str__ (self):
		return repr (self.ourValue)

class RepeatVariable (ContextVariable):
	""" To be written"""
	def __init__ (self, sequence):
		ContextVariable.__init__ (self, 1)
		self.sequence = sequence
		self.position = 0
		self.map = None

	def __getattr__(self, name): # to be used from python eval
 		if (self.map is None):
 			self.createMap()

 		return self.map[name]

	def value (self, currentPath=None):
		if (self.map is None):
			self.createMap()
		return self.map

	def rawValue (self):
		return self.value()

	def getCurrentValue (self):
		return self.sequence [self.position]

	def increment (self):
		self.position += 1
		if (self.position == len (self.sequence)):
			raise IndexError ("Repeat Finished")

	def createMap (self):
		self.map = {}
		self.map ['index'] = self.getIndex
		self.map ['number'] = self.getNumber
		self.map ['even'] = self.getEven
		self.map ['odd'] = self.getOdd
		self.map ['start'] = self.getStart
		self.map ['end'] = self.getEnd
		# TODO: first and last need to be implemented.
		self.map ['length'] = len (self.sequence)
		self.map ['letter'] = self.getLowerLetter
		self.map ['Letter'] = self.getUpperLetter
		self.map ['roman'] = self.getLowerRoman
		self.map ['Roman'] = self.getUpperRoman

	# Repeat implementation goes here
	def getIndex (self):
		return self.position

	def getNumber (self):
		return self.position + 1

	def getEven (self):
		if ((self.position % 2) != 0):
			return 0
		return 1

	def getOdd (self):
		if ((self.position % 2) == 0):
			return 0
		return 1

	def getStart (self):
		if (self.position == 0):
			return 1
		return 0

	def getEnd (self):
		if (self.position == len (self.sequence) - 1):
			return 1
		return 0

	def getLowerLetter (self):
		result = ""
		nextCol = self.position
		if (nextCol == 0):
			return 'a'
		while (nextCol > 0):
			nextCol, thisCol = divmod (nextCol, 26)
			result = chr (ord ('a') + thisCol) + result
		return result

	def getUpperLetter (self):
		return self.getLowerLetter().upper()

	def getLowerRoman (self):
		romanNumeralList = (('m', 1000)
						   ,('cm', 900)
						   ,('d', 500)
						   ,('cd', 400)
						   ,('c', 100)
						   ,('xc', 90)
						   ,('l', 50)
						   ,('xl', 40)
						   ,('x', 10)
						   ,('ix', 9)
						   ,('v', 5)
						   ,('iv', 4)
						   ,('i', 1)
						   )
		if (self.position > 3999):
			# Roman numbers only supported up to 4000
			return ' '
		num = self.position + 1
		result = ""
		for roman, integer in romanNumeralList:
			while (num >= integer):
				result += roman
				num -= integer
		return result

	def getUpperRoman (self):
		return self.getLowerRoman().upper()

class IteratorRepeatVariable (RepeatVariable):
	def __init__ (self, sequence):
		RepeatVariable.__init__ (self, sequence)
		self.curValue = None
		self.iterStatus = 0

	def getCurrentValue (self):
		if (self.iterStatus == 0):
			self.iterStatus = 1
			try:
				self.curValue = next(self.sequence)
			except StopIteration as e:
				self.iterStatus = 2
				raise IndexError ("Repeat Finished")
		return self.curValue

	def increment (self):
		# Need this for the repeat variable functions.
		self.position += 1
		try:
			self.curValue = next(self.sequence)
		except StopIteration as e:
			self.iterStatus = 2
			raise IndexError ("Repeat Finished")

	def createMap (self):
		self.map = {}
		self.map ['index'] = self.getIndex
		self.map ['number'] = self.getNumber
		self.map ['even'] = self.getEven
		self.map ['odd'] = self.getOdd
		self.map ['start'] = self.getStart
		self.map ['end'] = self.getEnd
		# TODO: first and last need to be implemented.
		self.map ['length'] = sys.maxsize
		self.map ['letter'] = self.getLowerLetter
		self.map ['Letter'] = self.getUpperLetter
		self.map ['roman'] = self.getLowerRoman
		self.map ['Roman'] = self.getUpperRoman

	def getEnd (self):
		if (self.iterStatus == 2):
			return 1
		return 0

class PathFunctionVariable (ContextVariable):
	def __init__ (self, func):
		ContextVariable.__init__ (self, value = func)
		self.func = func

	def value (self, currentPath=None):
		if (currentPath is not None):
			index, paths = currentPath
			result = ContextVariable (self.func ('/'.join (paths[index:])))
			# Fast track the result
			raise result

class CachedFuncResult (ContextVariable):
	def value (self, currentPath=None):
		try:
			return self.cachedValue
		except:
			self.cachedValue = ContextVariable.value (self)
		return self.cachedValue

	def clearCache (self):
		try:
			del self.cachedValue
		except:
			pass

class PythonPathFunctions:
	def __init__ (self, context):
		self.context = context

	def path (self, expr):
		return self.context.evaluatePath (expr)

	def string (self, expr):
		return self.context.evaluateString (expr)

	def exists (self, expr):
		return self.context.evaluateExists (expr)

	def nocall (self, expr):
		return self.context.evaluateNoCall (expr)

	def test (self, *arguments):
		if (len (arguments) % 2):
			# We have an odd number of arguments - which means the last one is a default
			pairs = arguments[:-1]
			defaultValue = arguments[-1]
		else:
			# No default - so use None
			pairs = arguments
			defaultValue = None

		index = 0
		while (index < len (pairs)):
			test = pairs[index]
			index += 1
			value = pairs[index]
			index += 1
			if (test):
				return value

		return defaultValue

class Context:
	def __init__ (self, options=None, allowPythonPath=0):
		self.allowPythonPath = allowPythonPath
		self.globals = {}
		self.locals = {}
		self.localStack = []
		self.repeatStack = []
		self.populateDefaultVariables (options)
		self.log = logging.getLogger ("simpleTALES.Context")
		self.true = 1
		self.false = 0
		self.pythonPathFuncs = PythonPathFunctions (self)

	def addRepeat (self, name, var, initialValue):
		# Pop the current repeat map onto the stack
		self.repeatStack.append (self.repeatMap)
		self.repeatMap = self.repeatMap.copy()
		self.repeatMap [name] = var
		# Map this repeatMap into the global space
		self.addGlobal ('repeat', self.repeatMap)

		# Add in the locals
		self.pushLocals()
		self.setLocal (name, initialValue)

	def removeRepeat (self, name):
		# Bring the old repeat map back
		self.repeatMap = self.repeatStack.pop()
		# Map this repeatMap into the global space
		self.addGlobal ('repeat', self.repeatMap)

	def addGlobal (self, name, value):
		self.globals[name] = value

	def pushLocals (self):
		# Push the current locals onto a stack so that we can safely over-ride them.
		self.localStack.append (self.locals)
		self.locals = self.locals.copy()

	def setLocal (self, name, value):
		# Override the current local if present with the new one
		self.locals [name] = value

	def popLocals (self):
		self.locals = self.localStack.pop()

	def evaluate (self, expr, originalAtts = None):
		# Returns a ContextVariable
		#self.log.debug ("Evaluating %s" % expr)
		if (originalAtts is not None):
			# Call from outside
			self.globals['attrs'] = originalAtts
			suppressException = 1
		else:
			suppressException = 0

		# Supports path, exists, nocall, not, and string
		expr = expr.strip ()
		try:
			if expr.startswith ('path:'):
				return self.evaluatePath (expr[5:].lstrip ())
			elif expr.startswith ('exists:'):
				return self.evaluateExists (expr[7:].lstrip())
			elif expr.startswith ('nocall:'):
				return self.evaluateNoCall (expr[7:].lstrip())
			elif expr.startswith ('not:'):
				return self.evaluateNot (expr[4:].lstrip())
			elif expr.startswith ('string:'):
				return self.evaluateString (expr[7:].lstrip())
			elif expr.startswith ('python:'):
				return self.evaluatePython (expr[7:].lstrip())
			else:
				# Not specified - so it's a path
				return self.evaluatePath (expr)
		except PathNotFoundException as e:
			if (suppressException):
				return None
			raise e

	def evaluatePython (self, expr):
		if (not self.allowPythonPath):
			self.log.warning ("Parameter allowPythonPath is false.  NOT Evaluating python expression %s" % expr)
			return self.false
		#self.log.debug ("Evaluating python expression %s" % expr)

		globals={}
		for name, value in list(self.globals.items()):
			if (isinstance (value, ContextVariable)): value = value.rawValue()
			globals [name] = value
		globals ['path'] = self.pythonPathFuncs.path
		globals ['string'] = self.pythonPathFuncs.string
		globals ['exists'] = self.pythonPathFuncs.exists
		globals ['nocall'] = self.pythonPathFuncs.nocall
		globals ['test'] = self.pythonPathFuncs.test

		locals={}
		for name, value in list(self.locals.items()):
			if (isinstance (value, ContextVariable)): value = value.rawValue()
			locals [name] = value

		try:
			result = eval(expr, globals, locals)
			if (isinstance (result, ContextVariable)):
				return result.value()
			return result
		except Exception as e:
			# An exception occured evaluating the template, return the exception as text
			self.log.warning ("Exception occurred evaluating python path, exception: " + str (e))
			return "Exception: %s" % str (e)

	def evaluatePath (self, expr):
		#self.log.debug ("Evaluating path expression %s" % expr)
		allPaths = expr.split ('|')
		if (len (allPaths) > 1):
			for path in allPaths:
				# Evaluate this path
				try:
					return self.evaluate (path.strip ())
				except PathNotFoundException as e:
					# Path didn't exist, try the next one
					pass
			# No paths evaluated - raise exception.
			raise PATHNOTFOUNDEXCEPTION
		else:
			# A single path - so let's evaluate it.
			# This *can* raise PathNotFoundException
			return self.traversePath (allPaths[0])

	def evaluateExists (self, expr):
		#self.log.debug ("Evaluating %s to see if it exists" % expr)
		allPaths = expr.split ('|')
		# The first path is for us
		# Return true if this first bit evaluates, otherwise test the rest
		try:
			result = self.traversePath (allPaths[0], canCall = 0)
			return self.true
		except PathNotFoundException as e:
			# Look at the rest of the paths.
			pass

		for path in allPaths[1:]:
			# Evaluate this path
			try:
				pathResult = self.evaluate (path.strip ())
				# If this is part of a "exists: path1 | exists: path2" path then we need to look at the actual result.
				if (pathResult):
					return self.true
			except PathNotFoundException as e:
				pass
		# If we get this far then there are *no* paths that exist.
		return self.false

	def evaluateNoCall (self, expr):
		#self.log.debug ("Evaluating %s using nocall" % expr)
		allPaths = expr.split ('|')
		# The first path is for us
		try:
			return self.traversePath (allPaths[0], canCall = 0)
		except PathNotFoundException as e:
			# Try the rest of the paths.
			pass

		for path in allPaths[1:]:
			# Evaluate this path
			try:
				return self.evaluate (path.strip ())
			except PathNotFoundException as e:
				pass
		# No path evaluated - raise error
		raise PATHNOTFOUNDEXCEPTION

	def evaluateNot (self, expr):
		#self.log.debug ("Evaluating NOT value of %s" % expr)

		# Evaluate what I was passed
		try:
			pathResult = self.evaluate (expr)
		except PathNotFoundException as e:
			# In SimpleTAL the result of "not: no/such/path" should be TRUE not FALSE.
			return self.true

		if (pathResult is None):
			# Value was Nothing
			return self.true
		if (pathResult == DEFAULTVALUE):
			return self.false
		try:
			resultLen = len (pathResult)
			if (resultLen > 0):
				return self.false
			else:
				return self.true
		except:
			# Not a sequence object.
			pass
		if (not pathResult):
			return self.true
		# Everything else is true, so we return false!
		return self.false

	def evaluateString (self, expr):
		#self.log.debug ("Evaluating String %s" % expr)
		result = ""
		skipCount = 0
		for position in range (0,len (expr)):
			if (skipCount > 0):
				skipCount -= 1
			else:
				if (expr[position] == '$'):
					try:
						if (expr[position + 1] == '$'):
							# Escaped $ sign
							result += '$'
							skipCount = 1
						elif (expr[position + 1] == '{'):
							# Looking for a path!
							endPos = expr.find ('}', position + 1)
							if (endPos > 0):
								path = expr[position + 2:endPos]
								# Evaluate the path - missing paths raise exceptions as normal.
								try:
									pathResult = self.evaluate (path)
								except PathNotFoundException as e:
									# This part of the path didn't evaluate to anything - leave blank
									pathResult = ''
								if (pathResult is not None):
									if (isinstance (pathResult, str)):
										result += pathResult
									else:
										# THIS IS NOT A BUG!
										# Use Unicode in Context if you aren't using Ascii!
										result += str (pathResult)
								skipCount = endPos - position
						else:
							# It's a variable
							endPos = expr.find (' ', position + 1)
							if (endPos == -1):
								endPos = len (expr)
							path = expr [position + 1:endPos]
							# Evaluate the variable - missing paths raise exceptions as normal.
							try:
								pathResult = self.traversePath (path)
							except PathNotFoundException as e:
								# This part of the path didn't evaluate to anything - leave blank
								pathResult = ''
							if (pathResult is not None):
								if (isinstance (pathResult, str)):
										result += pathResult
								else:
									# THIS IS NOT A BUG!
									# Use Unicode in Context if you aren't using Ascii!
									result += str (pathResult)
							skipCount = endPos - position - 1
					except IndexError as e:
						# Trailing $ sign - just suppress it
						self.log.warning ("Trailing $ detected")
						pass
				else:
					result += expr[position]
		return result

	def traversePath (self, expr, canCall=1):
		# canCall only applies to the *final* path destination, not points down the path.
		# Check for and correct for trailing/leading quotes
		if (expr.startswith ('"') or expr.startswith ("'")):
			if (expr.endswith ('"') or expr.endswith ("'")):
				expr = expr [1:-1]
			else:
				expr = expr [1:]
		elif (expr.endswith ('"') or expr.endswith ("'")):
			expr = expr [0:-1]
		pathList = expr.split ('/')

		path = pathList[0]
		if path.startswith ('?'):
			path = path[1:]
			if path in self.locals:
				path = self.locals[path]
				if (isinstance (path, ContextVariable)): path = path.value()
				elif (hasattr (path, "__call__")):path = path()

			elif path in self.globals:
				path = self.globals[path]
				if (isinstance (path, ContextVariable)): path = path.value()
				elif (hasattr (path, "__call__")):path = path()
				#self.log.debug ("Dereferenced to %s" % path)
		if path in self.locals:
			val = self.locals[path]
		elif path in self.globals:
			val = self.globals[path]
		else:
			# If we can't find it then raise an exception
			raise PATHNOTFOUNDEXCEPTION
		index = 1
		for path in pathList[1:]:
			#self.log.debug ("Looking for path element %s" % path)
			if path.startswith ('?'):
				path = path[1:]
				if path in self.locals:
					path = self.locals[path]
					if (isinstance (path, ContextVariable)): path = path.value()
					elif (hasattr (path, "__call__")):path = path()
				elif path in self.globals:
					path = self.globals[path]
					if (isinstance (path, ContextVariable)): path = path.value()
					elif (hasattr (path, "__call__")):path = path()
				#self.log.debug ("Dereferenced to %s" % path)
			try:
				if (isinstance (val, ContextVariable)): temp = val.value((index,pathList))
				elif (hasattr (val, "__call__")):temp = val()
				else: temp = val
			except ContextVariable as e:
				# Fast path for those functions that return values
				return e.value()

			if (hasattr (temp, path)):
				val = getattr (temp, path)
			else:
				try:
					try:
						val = temp[path]
					except TypeError:
						val = temp[int(path)]
				except:
					#self.log.debug ("Not found.")
					raise PATHNOTFOUNDEXCEPTION
			index = index + 1
		#self.log.debug ("Found value %s" % str (val))
		if (canCall):
			try:
				if (isinstance (val, ContextVariable)): result = val.value((index,pathList))
				elif (hasattr (val, "__call__")):result = val()
				else: result = val
			except ContextVariable as e:
				# Fast path for those functions that return values
				return e.value()
		else:
			if (isinstance (val, ContextVariable)): result = val.realValue
			else: result = val
		return result

	def __str__ (self):
		return "Globals: " + str (self.globals) + "Locals: " + str (self.locals)

	def populateDefaultVariables (self, options):
		vars = {}
		self.repeatMap = {}
		vars['nothing'] = None
		vars['default'] = DEFAULTVALUE
		vars['options'] = options
		# To start with there are no repeats
		vars['repeat'] = self.repeatMap
		vars['attrs'] = None

		# Add all of these to the global context
		for name in list(vars.keys()):
			self.addGlobal (name,vars[name])

		# Add also under CONTEXTS
		self.addGlobal ('CONTEXTS', vars)



#!/usr/bin/env python

"""
embed package

To use the embed package, include the following code in your latex document.
It will define the HTML environment and the \html command.  The HTML
environment is exactly the same as the comment environment.  The \html
command acts like the \verb command in that it will eat up all content
between two matching characters (e.g., \html+<b>bold text</b>+).  In
plasTeX, the content of both of these macros will appear in the output
document as-is.

\newif\ifplastex\plastexfalse
\ifplastex
    \\usepackage{embed}
\else
    \\usepackage{comment}\excludecomment{HTML}
    \def\html#1{\def\@html##1#1{}\@html}
\fi

Example:

before \html+<span style="color:red">red text</span>+ after

before
\begin{HTML}
<div style="background-color:blue; color:white">
<p>line 1</p>
<p><b>line 2</b></p>
</div>
\end{HTML}
after

"""

from plasTeX import Command, Environment
from plasTeX.DOM import Text
from plasTeX.Base.LaTeX.Verbatim import verbatim
from plasTeX.Base.LaTeX.Verbatim import verb

class HTML(verbatim):
    captionable = True
    blockType = False
    
    def digest(self, tokens):
        verbatim.digest(self, tokens)
        self.str = Text(''.join(self))
        self.str.isMarkup = True
        return []
    
class XHTML(HTML):
    pass
    
class html(verb):
    args = ''
    
    def digest(self, tokens):
        verb.digest(self, tokens)
        self.str = Text(''.join(self))
        self.str.isMarkup = True
        return []
    
class xhtml(html):
    pass


""" Fixed up HTMLParser

		Copyright (c) 2009 Colin Stewart (http://www.owlfish.com/)
		All rights reserved.

		Redistribution and use in source and binary forms, with or without
		modification, are permitted provided that the following conditions
		are met:
		1. Redistributions of source code must retain the above copyright
		   notice, this list of conditions and the following disclaimer.
		2. Redistributions in binary form must reproduce the above copyright
		   notice, this list of conditions and the following disclaimer in the
		   documentation and/or other materials provided with the distribution.
		3. The name of the author may not be used to endorse or promote products
		   derived from this software without specific prior written permission.

		THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
		IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
		OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
		IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
		INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
		NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
		DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
		THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
		(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
		THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

		If you make any bug fixes or feature enhancements please let me know!


		The classes in this module implement the TAL language, expanding
		both XML and HTML templates.

		Module Dependencies: logging, simpleTALES, simpleTALTemplates
"""

from .. import simpletal
import html.parser


class HTMLParser (html.parser.HTMLParser):
	def unescape(self, s):
		# Just return the data - we don't partially unescaped data!
		return s



"""
Implement the tikz package for html output.
The original tikz and latex do most of the work and then pdf2svg (or similar
software) turns it into a svg image.
Needs Beautiful Soup, Jinja2, and pdf2svg or similar
"""

import os
import string
import subprocess
import shutil
import tempfile
from plasTeX import Environment, NoCharSubEnvironment, Macro
from plasTeX.PackageResource import PackageResource

from plasTeX.Logging import getLogger
log = getLogger()

try:
    from jinja2 import Template
except ImportError:
    log.warning('Cannot find jinja2 lib. Cannot use tikz.')

try:
    from bs4 import BeautifulSoup
except ImportError:
    log.warning('Cannot find BeautifulSoup lib. Cannot use tikz.')


class tikzpicture(NoCharSubEnvironment):
    """
    A tikz picture whose content will be converted in the processFileContent callback.
    """
    class matrix(Environment):
        """
        Avoids conflict with amsmath matrix thanks to the context stack
        mechanism.
        """

    class draw(Macro):
        """ Only avoids unrecognized command warning. """

    class fill(Macro):
        """ Only avoids unrecognized command warning. """

    class filldraw(Macro):
        """ Only avoids unrecognized command warning. """

    class node(Macro):
        """ Only avoids unrecognized command warning. """

    class path(Macro):
        """ Only avoids unrecognized command warning. """

    class clip(Macro):
        """ Only avoids unrecognized command warning. """

def tikzConvert(document, content, envname, placeholder):
    tmp_dir = document.userdata[envname]['tmp_dir']
    working_dir = document.userdata['working-dir']
    outdir = document.config['files']['directory']
    outdir = string.Template(outdir).substitute(
            {'jobname': document.userdata.get('jobname', '')})
    target_dir = os.path.join(working_dir, outdir, 'images')
    if not os.path.isdir(target_dir):
        os.makedirs(target_dir)
    template = document.userdata[envname]['template']
    compiler = document.userdata[envname]['compiler']
    pdf2svg = document.userdata[envname]['pdf2svg']

    cwd = os.getcwd()
    os.chdir(tmp_dir)
    soup = BeautifulSoup(content, "html.parser")
    encoding = soup.original_encoding

    envs = soup.findAll(envname)
    for env in envs:
        object_id = env.attrs['id']
        basepath = os.path.join(tmp_dir, object_id)
        texpath = basepath + '.tex'
        pdfpath = basepath + '.pdf'
        svgpath =  basepath + '.svg'

        context = { 'tikzpicture': env.text.strip() }
        template.stream(**context).dump(texpath, encoding)

        subprocess.call([compiler, texpath])
        subprocess.call([pdf2svg, pdfpath, svgpath])
        destination = os.path.join(target_dir, object_id+'.svg')
        if os.path.isfile(destination):
            os.remove(destination)
        shutil.move(svgpath, target_dir)

        obj = soup.new_tag(
                'object',
                type='image/svg+xml',
                data='images/' + object_id + '.svg')
        obj.string = document.context.terms.get(
                placeholder,
                placeholder) + '\n' + env.text.strip()
        obj.attrs['class'] = envname
        div = soup.new_tag('div')
        div.attrs['class'] = envname
        div.insert(1, obj)

        env.replace_with(div)
    os.chdir(cwd)
    try:
        # python2
        result = unicode(soup)
    except NameError:
        # python3
        result = str(soup)
    return result

def ProcessOptions(options, document):
    """This is called when the package is loaded."""

    try:
        with open(document.config['html5']['tikz-template'], "r") as file:
            template = file.read()
    except IOError:
        log.info('Using default TikZ template.')
        template = u"\\documentclass{standalone}\n\\usepackage{tikz}" + \
                   u"\\begin{document}{{ tikzpicture }}\\end{document}"
    document.userdata['tikzpicture'] = {
            'template': Template(template),
            'tmp_dir': tempfile.mkdtemp(),
            'compiler': document.config['html5']['tikz-compiler'],
            'pdf2svg': document.config['html5']['tikz-converter'],
            }

    def convert(document, content):
        return tikzConvert(document, content, 'tikzpicture', 'TikZ picture')

    cb = PackageResource(
            renderers='html5',
            key='processFileContents',
            data=convert)
    document.addPackageResource(cb)

    cb = PackageResource(
            renderers='gerby',
            key='processFileContents',
            data=convert)
    document.addPackageResource(cb)



#!/usr/bin/env python

from plasTeX.Logging import getLogger
from plasTeX.Imagers import Imager as _Imager
import glob, sys

status = getLogger('status')

gs = 'gs'
if sys.platform.startswith('win'):
   gs = 'gswin32c'

class GSPDFPNG(_Imager):
    """ Imager that uses gs to convert pdf to png """
    command = ('%s -dSAFER -dBATCH -dNOPAUSE -sDEVICE=png16m -r250 ' % gs) + \
              '-dGraphicsAlphaBits=4 -sOutputFile=img%d.png'
    compiler = 'pdflatex'
    fileExtension = '.png'

    def executeConverter(self, output):
        res = plasTeX.Imagers.Imager.executeConverter(self, output)
        self.scaleImages()
        return res

    def scaleImages(self):
        """ Scale images down and anti-alias """
        if plasTeX.Imagers.PILImage is not None:
            PILImage = plasTeX.Imagers.PILImage
            scaledown = 2.2
            for filename in glob.glob('img*.png'):
                status.info('[%s]' % filename,)
                img = plasTeX.Imagers.autoCrop(PILImage.open(filename),
                                               margin=3)[0]
                width, height = [int(float(x)/scaledown) for x in img.size]
                img = img.resize((width, height), PILImage.ANTIALIAS)
                img = img.point(self.toWhite)
                img.save(filename)

    def toWhite(self, pixel):
        if pixel >= 245:
            return 255
        return pixel

    def formatConfigOptions(self, config):
        options = []
        if config['resolution']:
            options.append(('-r%s' % config['resolution'], ''))
        return options

Imager = GSPDFPNG



#!/usr/bin/env python

from plasTeX.Renderers import Renderer as BaseRenderer
from plasTeX import encoding
import textwrap, re, string

class TextRenderer(BaseRenderer):
    """ Renderer for plain text documents """

    outputType = str
    fileExtension = '.txt'
    lineWidth = 76

    aliases = {
        'superscript': 'active::^',
        'subscript': 'active::_',
        'dollar': '$',
        'percent': '%',
        'opencurly': '{',
        'closecurly': '}',
        'underscore': '_',
        'ampersand': '&',
        'hashmark': '#',
        'space': ' ',
        'tilde': 'active::~',
        'at': '@',
        'backslash': '\\',
    }

    def __init__(self, *args, **kwargs):
        BaseRenderer.__init__(self, *args, **kwargs)

        # Load dictionary with methods
        for key in dir(self):
            if key.startswith('do__'):
                self[self.aliases[key[4:]]] = getattr(self, key)
            elif key.startswith('do_'):
                self[key[3:]] = getattr(self, key)

        self['default-layout'] = self['document-layout'] = self.default

        self.footnotes = []
        self.blocks = []

    def addBlock(self, s):
        """ Create a new block level element placeholder """
        self.blocks.append(s)
        width = max([len(x) for x in s.split('\n')])
        return u'\001[%s%s]' % (len(self.blocks)-1, '@'*width)

    def default(self, node):
        """ Rendering method for all non-text nodes """
        # Handle characters like \&, \$, \%, etc.
        if len(node.nodeName) == 1 and node.nodeName not in encoding.stringletters():
            return self.textDefault(node.nodeName)

        # Render child nodes
        return str(node)

    def processFileContent(self, document, s):
        s = BaseRenderer.processFileContent(self, document, s)

        # Put block level elements back in
        block_re = re.compile('(\\s*)\001\\[(\\d+)@+\\]')
        while 1:
            m = block_re.search(s)
            if not m:
                break

            space = ''
            before = m.group(1)
            if before is None:
                before = ''
            if '\n' in before:
                spaces = before.split('\n')
                space = spaces.pop()
                before = '\n'.join(spaces) + '\n'

            block = self.blocks[int(m.group(2))]
            block = space + block.replace('\n', u'\n%s' % space)

            s = block_re.sub('%s%s' % (before, block), s, 1)

        # Clean up newlines
        return re.sub(r'\s*\n\s*\n(\s*\n)+', r'\n\n\n', s)

    def textDefault(self, node):
        return str(node)

    def wrap(self, s, **kwargs):
        return textwrap.wrap(str(s), self.lineWidth, break_long_words=False, **kwargs)

    def fill(self, s, **kwargs):
        return textwrap.fill(str(s), self.lineWidth, break_long_words=False, **kwargs)

    # Alignment

    def do_flushleft(self, node):
        return self.fill(str(node)).strip()

    do_raggedbottom = do_raggedright = do_leftline = do_flushleft

    def center(self, text):
        s = self.wrap(text)
        for i, line in enumerate(s):
            indent = int((self.lineWidth - len(line)) / 2)
            s[i] = '%s%s' % (' '*indent, line)
        return '\n'.join(s)

    def do_center(self, node):
        return self.center(str(node))

    do_centering = do_centerline = do_center

    def do_flushright(self, node):
        s = self.wrap(node)
        for i, line in enumerate(s):
            indent = self.lineWidth - len(line)
            s[i] = '%s%s' % (' '*indent, line)
        return '\n'.join(s).rstrip()

    do_raggedleft = do_llap = do_flushright

    # Arrays

    def do_array(self, node, cellspacing=(2,1), render=str):
        # Render the table cells and get min and max column widths
        colwidths = []
        for r, row in enumerate(node):
            for c, cell in enumerate(row):
                if isinstance(render, str):
                    s = getattr(cell, render)().strip()
                else:
                    s = render(cell).strip()
                if s.strip():
                    maxlength = max([len(x) for x in s.split('\n')])
                    minlength = min([len(x) for x in s.split() if x.strip()])
                else:
                    minlength = maxlength = 0
                if r == 0:
                    colwidths.append([minlength, maxlength])
                else:
                    colwidths[c] = [max(minlength, colwidths[c][0]),
                                    max(maxlength, colwidths[c][1])]

        # Determine best column widths
        maxline = self.lineWidth - len(colwidths)*cellspacing[0]
        minwidths = [x[0] for x in colwidths] # minimums
        maxwidths = [x[1] for x in colwidths] # maximums
        if sum(maxwidths) < maxline:
            outwidths = maxwidths
        elif sum(minwidths) > maxline:
            outwidths = minwidths
        else:
            outwidths = list(maxwidths)
            # If the minimum is also the maximum, take it out of the
            # algorithm to determine lengths.
            for i, item in enumerate(maxwidths):
                if maxwidths[i] == minwidths[i]:
                    maxwidths[i] = -1
            # Iteratively subtract one from the longest line until the
            # table will fit within maxline.
            while sum(outwidths) > maxline:
                index = maxwidths.index(max(maxwidths))
                maxwidths[i] -= 1
                outwidths[i] -= 1
                if maxwidths[i] == minwidths[i]:
                    maxwidths[i] = -1

        # Render cells to correct widths
        rendered = []
        for r, row in enumerate(node):
            current = []
            rendered.append(current)
            for c, cell in enumerate(row):
                origwidth = self.lineWidth
                self.lineWidth = outwidths[c]
                if isinstance(render, str):
                    s = getattr(cell, render)().split('\n')
                else:
                    s = render(cell).strip().split('\n')
                if s and r == 0 and c == 0:
                    s[0] = s[0].lstrip()
                current.append((max([len(x) for x in s]), len(s), s))
                self.lineWidth = origwidth

        # Pad cells to fill out a block
        for r, row in enumerate(rendered):
            linesneeded = max([x[1] for x in row]) + cellspacing[1]
            for c, cell in enumerate(row):
                width, height, content = cell
                # Add the appropriate number of lines
                for i in range(linesneeded - len(content)):
                    content.append(' '*width)
                # Pad all lines to the same length
                for i, line in enumerate(content):
                    content[i] = content[i] + ' '*(outwidths[c]-width+cellspacing[0])
                rendered[r][c] = content

        # Piece all of the table parts together
        output = []
        for row in rendered:
            # Continue until cell content is empty (the last cell in this case)
            while row[-1]:
                for cell in row:
                    output.append(cell.pop(0))
                # Get rid of unneeded whitespace and put in a line break
                output[-1] = output[-1].strip()
                output.append('\n')

        return ''.join(output)

    do_tabular = do_tabularx = do_longtable = do_array

    def do_cline(self, node):
        return ''

    def do_multicolumn(self, node):
        return str(node)

    # Bibliography
    def do_thebibliography(self, node):
        output = ['', 'Bibliography', '']
        for item in node:
            bullet = u'[%s] ' % item.bibcite
            bulletlen = len(bullet)
            output.append(self.fill(item, initial_indent=bullet,
                                        subsequent_indent=' '*bulletlen))
            output.append('')
        output.append('')
        return u'\n'.join(output)

    def do_bibliographystyle(self, node):
        return u''

    def do_bibliography(self, node):
        return self.default(node)

    def do_cite(self, node):
        output = []
        for item in node.citation():
            output.append(str(item))
        return u''.join(output)

    def do_bibliographyref(self, node):
        return self.default(node)

    # Boxes

    do_mbax = do_makebox = do_fbox = do_framebox = do_parbox = default
    do_minipage = do_raisebox = do_rule = default

    # Breaking

    def do_linebreak(self, node):
        return u'\n\n'

    do_newline = do_pagebreak = do_newpage = do_clearpage = do_cleardoublepage = do_linebreak

    # Crossref

    def do_ref(self, node):
        return str(node.idref['label'].ref)

    def do_pageref(self, node):
        return u'*'

    def do_label(self, node):
        return u''

    # Floats

    def do_figure(self, node):
        return str(node)

    do_table = do_marginpar = do_figure

    def do_caption(self, node):
        return u'%s %s: %s' % (node.title, node.ref, str(node))

    # Font Selection

    do_mdseries = do_textmd = do_bfseries = do_textbf = default
    do_rmfamily = do_textrm = do_sffamily = do_textsf = default
    do_ttfamily = do_texttt = do_upshape = do_textup = default
    do_itshape = do_textit = do_scshape = do_textsc = default
    do_slshape = do_textsl = do_textnormal = do_rm = default
    do_cal = do_it = do_sl = do_bf = do_tt = do_sc = default
    do_tiny = do_scriptsize = do_footnotesize = do_small = default
    do_normalsize = do_large = do_Large = do_LARGE = do_huge = do_HUGE = default

    def do_symbol(self, node):
        return u'*'

    # Footnotes
    def do_footnote(self, node):
        mark = u'[%s]' % (len(self.footnotes)+1)
        self.footnotes.append(self.fill(node, initial_indent='%s ' % mark,
                                     subsequent_indent=' ' * (len(mark)+1)).strip())
        return mark

    def do_footnotetext(self, node):
        self.do_footnote(self, node)
        return ''

    def do_footnotemark(self, node):
        return u'[%s]' % (len(self.footnotes)+1)

    # Index

    def do_theindex(self, node):
        return u''

    do_printindex = do_index = do_theindex

    # Lists

    def do_itemize(self, node):
        output = []
        for item in node:
            bullet = '   * '
            bulletlen = len(bullet)
            output.append(self.fill(item, initial_indent=bullet,
                                        subsequent_indent=' '*bulletlen))
        return self.addBlock(u'\n'.join(output))

    def do_enumerate(self, node):
        output = []
        for i, item in enumerate(node):
            bullet = '   %d. ' % (i+1)
            bulletlen = len(bullet)
            output.append(self.fill(item, initial_indent=bullet,
                                        subsequent_indent=' '*bulletlen))
        return self.addBlock(u'\n'.join(output))

    def do_description(self, node):
        output = []
        for item in node:
            bullet = '   %s - ' % item.attributes.get('term','')
            bulletlen = len(bullet)
            output.append(self.fill(item, initial_indent=bullet,
                                        subsequent_indent='      '))
        return self.addBlock(u'\n'.join(output))

    do_list = do_trivlist = do_description

    # Math

    def do_math(self, node):
        return re.sub(r'\s*(_|\^)\s*', r'\1', node.source)

    do_ensuremath = do_math

    def do_equation(self, node):
        s = u'   %s' % re.compile(r'^\s*\S+\s*(.*?)\s*\S+\s*$', re.S).sub(r'\1', node.source)
        return re.sub(r'\s*(_|\^)\s*', r'\1', s)

    do_displaymath = do_equation

    def do_eqnarray(self, node):
        def render(node):
            s = re.compile(r'^\$\\displaystyle\s*(.*?)\s*\$\s*$', re.S).sub(r'\1', node.source)
            return re.sub(r'\s*(_|\^)\s*', r'\1', s)
        s = self.do_array(node, cellspacing=(1,1), render=render)
        output = []
        for line in s.split('\n'):
            output.append('   %s' % line)
        return u'\n'.join(output)

    do_align = do_gather = do_falign = do_multiline = do_eqnarray
    do_multline = do_alignat = do_split = do_eqnarray

    # Misc

    do_bgroup = default

    def do_def(self, node):
        return u''

    do_tableofcontents = do_input = do_protect = do_let = do_def
    do_newcommand = do_hfill = do_hline = do_openout = do_renewcommand = do_def
    do_write = do_hspace = do_appendix = do_global = do_noindent = do_def
    do_include = do_markboth = do_setcounter = do_refstepcounter = do_def
    do_medskip = do_smallskip = do_parindent = do_indent = do_setlength = do_def
    do_settowidth = do_addtolength = do_nopagebreak = do_newwrite = do_def
    do_newcounter = do_typeout = do_sloppypar = do_hfil = do_thispagestyle = do_def

    def do_egroup(self, node):
        return u''

    # Pictures

    def do_picture(self, node):
        return u''

    # Primitives

    def do_par(self, node):
        numchildren = len(node.childNodes)
        if numchildren == 1 and not isinstance(node[0], str):
            return u'%s\n\n' % str(node)
        elif numchildren == 2 and isinstance(node[1], str) and \
           not node[1].strip():
            return u'%s\n\n' % str(node)
        s = u'%s\n\n' % self.fill(node)
        if not s.strip():
            return u''
        return s

    def do__superscript(self, node):
        return self.default(node)

    def do__subscript(self, node):
        return self.default(node)

    # Quotations

    def do_quote(self, node):
        backslash = self['\\']
        self['\\'] = lambda *args: u'\001'
        res = [x.strip() for x in str(node).split(u'\001')]
        output = []
        for par in [x.strip() for x in str(node).split(u'\n\n')]:
            for item in [x.strip() for x in par.split(u'\001')]:
                output.append(self.fill(item, initial_indent='   ', subsequent_indent='      '))
            output.append('')
        output.pop()
        self['\\'] = backslash
        return u'\n'.join(output)

    do_quotation = do_verse = do_quote

    # Sectioning

    def do_document(self, node):
        content = str(node).rstrip()
        footnotes = u'\n\n'.join(self.footnotes).rstrip()
        if footnotes:
            content = u'%s\n\n\n%s' % (content, footnotes)
        return u'%s\n\n' % content

    def do_maketitle(self, node):
        output = []
        metadata = node.ownerDocument.userdata
        if 'title' in metadata:
            output.append(self.center(metadata['title']).rstrip().upper())
            output.append('')
        if 'author' in metadata:
            for author in metadata['author']:
                if [a for a in author if getattr(a,'macroName','') == '\\']:
                    for a in author:
                        if getattr(a,'macroName','') == '\\':
                            continue
                        output.append(self.center(a).rstrip())
                else:
                    output.append(self.center(author).rstrip())
            output.append('')
        if 'date' in metadata:
            output.append(self.center(metadata['date']).rstrip())
            output.append('')
        if 'thanks' in metadata:
            output.append(self.center(metadata['thanks']).rstrip())
        return u'\n%s\n\n' % u'\n'.join(output)

    def do_section(self, node):
        return u'\n\n\n%s' % (u'%s\n\n%s' % (self.fill(node.fullTitle), node)).strip()

    do_part = do_chapter = do_subsection = do_subsubsection = do_section
    do_paragraph = do_subparagraph = do_subsubparagraph = do_section

    def do_title(self, node):
        return u''

    do_author = do_date = do_thanks = do_title

    def do_abstract(self, node):
        return self.center(node)

    # Sentences

    def do__dollar(self, node):
        return u'$'

    def do__percent(self, node):
        return u'%'

    def do__opencurly(self, node):
        return u'{'

    def do__closecurly(self, node):
        return u'}'

    def do__underscore(self, node):
        return u'_'

    def do__ampersand(self, node):
        return u'&'

    def do__hashmark(self, node):
        return u'#'

    def do__space(self, node):
        return u' '

    def do_LaTeX(self, node):
        return u'LaTeX'

    def do_TeX(self, node):
        return u'TeX'

    def do_emph(self, node):
        return self.default(node)

    do_em = do_emph

    def do__tilde(self, node):
        return u' '

    def do_enspace(self, node):
        return u' '

    do_quad = do_qquad = do_enspace

    def do_enskip(self, node):
        return u''

    do_thinspace = do_enskip

    def do_underbar(self, node):
        return self.default(node)

    # Space

    def do_hspace(self, node):
        return u' '

    def do_vspace(self, node):
        return u''

    do_bigskip = do_medskip = do_smallskip = do_vspace

    # Tabbing - not implemented yet

    # Verbatim

    def do_verbatim(self, node):
        return re.sub(r'^\s*\n', r'', str(node)).rstrip()

    do_alltt = do_verbatim

    def do_mbox(self, node):
        return self.default(node)

    def do__at(self, node):
        return u''

    def do__backslash(self, node):
        return u'\\'

Renderer = TextRenderer



#!/usr/bin/env python

import subprocess, shlex
import os, shutil, re
from plasTeX.Renderers.PageTemplate import Renderer as _Renderer
from plasTeX.Renderers import Renderer as BaseRenderer
from plasTeX.Logging import getLogger

log = getLogger()

class HTML5(_Renderer):
    """ Renderer for HTML5 documents, heavily copied from XHTML renderer """

    fileExtension = '.html'
    imageTypes = ['.svg', '.png','.jpg','.jpeg','.gif']
    vectorImageTypes = ['.svg']

    def loadTemplates(self, document):
        """Load templates as in PageTemplate but also look for packages that
        want to override some templates and handles extra css and javascript."""

        try:
            import jinja2
        except ImportError:
            log.error('Jinja2 is not available, hence the HTML5 renderer cannot be used.')

        _Renderer.loadTemplates(self, document)
        rendererdata = document.rendererdata['html5'] = dict()
        config = document.config

        rendererDir = os.path.dirname(__file__)

        srcDir = document.userdata['working-dir']
        buildDir = os.getcwd()

        # Theme css has already been copied by PageTemplate.loadTemplates,
        # provided config['general']['copy-theme-extras'] is true
        # Still try to create styles directory is case it's false
        try:
            os.mkdir('styles')
        except OSError:
            # This should mean the directory already exists
            pass

        # Start building the css list for use by the layout template
        if config['html5']['use-theme-css'] and config['general']['copy-theme-extras']:
            rendererdata['css'] = ['theme-' + config['html5']['theme-css'] + '.css']
        else:
            rendererdata['css'] = []

        # Theme js has already been copied by PageTemplate.loadTemplates,
        # provided config['general']['copy-theme-extras'] is true
        # Still try to create js directory is case it's false
        try:
            os.mkdir('js')
        except OSError:
            pass

        # Start building the js list for use by the layout template
        if (config['html5']['use-theme-js'] and 
                config['general']['copy-theme-extras']):
            rendererdata['js'] = sorted(
                    os.listdir(os.path.join(self.loadedTheme, 'js')))
        else:
            rendererdata['js'] = []

        for resrc in document.packageResources:
            # Next line may load templates or change
            # document.rendererdata['html5'] or copy some files to buildDir
            resrc.alter(
                    renderer=self,
                    rendererName='html5',
                    document=document,
                    target=buildDir)

        # Last loaded files (hence overriding everything else) come from user
        # configuration
        cssBuildDir = os.path.join(buildDir, 'styles')
        for css in config['html5']['extra-css']:
            rendererdata['css'].append(css)
            shutil.copy(os.path.join(srcDir, css), cssBuildDir)

        jsBuildDir = os.path.join(buildDir, 'js')
        for js in config['html5']['extra-js']:
            rendererdata['js'].append(js)
            shutil.copy(os.path.join(srcDir, js), jsBuildDir)

    def processFileContent(self, document, s):
        s = _Renderer.processFileContent(self, document, s)

        # Remove empty paragraphs
        s = re.compile(r'<p>\s*</p>', re.I).sub(r'', s)

        # Add a non-breaking space to empty table cells
        s = re.compile(r'(<(td|th)\b[^>]*>)\s*(</\2>)', re.I).sub(r'\1&nbsp;\3', s)

        for fun in document.rendererdata['html5'].get('processFileContents', []):
            s = fun(document, s)

        filters = document.config['html5']['filters']
        for filter_ in filters:
            proc = subprocess.Popen(
                    shlex.split(filter_),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE)
            output, output_err = proc.communicate(s.encode(encoding="utf-8"))
            if not output_err:
                s = output.decode(encoding="utf-8")

        return s

    def cleanup(self, document, files, postProcess=None):
        """
        Cleanup method called at the end of rendering.
        Uses the base renderer cleanup but calls packages callbacks before and
        after. Callbacks should be listed in
        document.userdata['preCleanupCallbacks']
        or document.userdata['postCleanupCallbacks']. Each call back should accept the
        current document as its only argument. Pre-cleanup call back must return
        the list of path of files they created (relative to the output directory).
        """

        rendererdata = document.rendererdata.get('html5', dict())
        preCleanupCallbacks = rendererdata.get('preCleanupCallbacks', [])
        for preCleanupCallback in preCleanupCallbacks:
            files += preCleanupCallback(document)

        _Renderer.cleanup(self, document, files, postProcess)


Renderer = HTML5



#!/usr/bin/env python

from plasTeX import Command, Environment

class category(Environment):
    args = '[ label ] title'

class SBentries(Command):
    args = ' entries:list '

class SBtitlestyle(Command):
    args = 'type:str'

class SBsubtitlestyle(Command):
    args = 'type:str'



#!/usr/bin/env python

import re, os, copy
from plasTeX.ConfigManager import BUILTIN, CODE, ENVIRON, ENVIRONMENT
from plasTeX.ConfigManager import CODE, REGISTRY, COMMANDLINE, InvalidOptionError
from plasTeX.ConfigManager import ConfigManager, TooManyValues, TooFewValues
import collections
collections.Callable = collections.abc.Callable

DEFAULTS = \
{
   'docstring': '',
   'options': None,
   'default': None,
   'optional': None,
   'values': None,
   'category': None,
   'name': None,
   'source': BUILTIN,
   'callback': None,
   'synopsis': None,
   'environ': None,
   'registry': None,
   'mandatory': None,
}

RegexType = type(re.compile(r''))


class GenericParser:

   def getOptionalArgument(self, args):
      return self.getArgument(args, range=[0,1])

   def _hasFollowingArgument(self, args, delim):
      """ Return boolean indicating the existence of another argument """
      if not ConfigManager.has_following_argument(args):
         return 0

      # No arguments left
      if not args:
         return 0

      # The current argument ends with the proper delimiter
      if args[0].strip()[-1] == delim:
         return 1

      # The next argument begins with the proper delimiter
      if len(args) > 1 and args[1].strip()[0] == delim:
         return 1

      return 0

   def getArgument(self, args, range=None, delim=',', forcedarg=False):
      """ Parse argument """

      range = self.validateRange(range or [1, 1])

      # If we have an optional default, make sure that the range reflects that
      if range[0] < 2 and self.optional not in [None, [], ()]:
         range[0] = 0

      # No arguments left
      if not args:
         # Arguments that aren't mandatory can bail out now.
         if isinstance(self, GenericArgument) and not(self.mandatory):
            return None, args

         if range[0] > 0:
            name = self.name
            if self.actual: name = self.actual
            if range[0] != range[1]:
               raise TooFewValues("Expecting at least %s value(s) for '%s'." % (range[0], name))
            elif range[0] == 1:
               raise TooFewValues("Expecting a value for '%s'." % name)
            else:
               raise TooFewValues("Expecting %s values for '%s'." % (range[0], name))
         if self.optional:
            return self.optional, args
         else:
            return None, args

      # If delimiter is whitespace...
      if not delim or not str(delim).strip():
         new_args, args = self._getSpaceDelimitedArguments(args, range)
         delim = ' '

      else:
         new_args = []
         while self._hasFollowingArgument(args, delim):
            forcedarg = False

            # If the current argument ends with a delimiter
            if args[0].strip()[-1] == delim:
               new_args.append(args.pop(0).strip()[:-1])

            # If next argument is an isolated delimiter character
            elif len(args) > 1 and args[1].strip() == delim:
               new_args.append(args.pop(0).strip())
               args.pop(0)

            # If next argument begins with a delimiter
            else:
               if args[0].startswith(delim):
                   new_args.append(args.pop(0).strip()[1:])
               else:
                   new_args.append(args.pop(0).strip())

               # If this argument ends with a delimiter, rip it off
               # and put it back into the stream.
               if new_args[-1].endswith(delim):
                  args.insert(0, delim)
                  new_args[-1] = new_args[-1][:-1]

         # The following argument is always part of the list
         # unless the user accidentally put a trailing delimiter.
         if forcedarg or ConfigManager.has_following_argument(args):
            new_args.append(args.pop(0).strip())

         if new_args and new_args[-1].startswith(delim):
            new_args[-1] = new_args[-1][1:]

      if len(range) == 2 and range[0] == 1 and range[1] == 1:
          pass
      else:
          _args = []
          for i in new_args:
             _args += i.split(delim)
          new_args = _args

      # Check number of values
      name = self.name
      if self.actual: name = self.actual
      if len(new_args) < range[0]:
         plural = 's'
         if range[0] == 1:
            plural = ''
         raise TooFewValues("Expecting at least %s value%s for '%s'." % (range[0], plural, name))
      elif len(new_args) > range[1]:
         raise TooManyValues("Expecting at most %s values for '%s'." % (range[1], name))

      # Collapse argument list to a value if possible
      if len(new_args) < 1:
         if self.optional is not None:
            new_args = self.optional
         else:
            new_args = None
      elif len(new_args) < 2:
         new_args = new_args[0]

      return new_args, args

   def _getSpaceDelimitedArguments(self, args, range=None):
      """ Slurp up any following arguments that don't begin with '-' """
      minimum, maximum = self.validateRange(range or [1, '*'])
      arg_len = len(args)
      new_args = []
      while ConfigManager.has_following_argument(args):
          new_args.append(args.pop(0))

      # If all of the remaining command line components are arguments as
      # opposed to options, don't return them as arguments to the
      # last option.
      if isinstance(self, GenericArgument):
          return new_args, args
      else:
          if len(new_args) == arg_len:
              minimum = min(len(new_args), minimum)
              return new_args[:minimum], new_args[minimum:]
          else:
              return new_args, args

   def validateRange(self, range):
      # Verify that ranges are valid
      if range[0] == '*': range[0] = 10000
      if range[1] == '*': range[1] = 10000
      range[0] = max(range[0], 0)
      if range[0] > range[1]:
         range[0] = range[1]
      range[0] = int(range[0])
      range[1] = int(range[1])
      return range


class GenericOption(object):
   """
   Base class for configuration option classes

   All command line options must subclass from this one.  In addition,
   this is an abstract class.  It cannot be instantiated directly.

   """

   sort_order = ['category', 'options']
   synopsis = ''

   def __init__(self, docstring=DEFAULTS['docstring'],
                      options=DEFAULTS['options'],
                      default=DEFAULTS['default'],
                      optional=DEFAULTS['optional'],
                      values=DEFAULTS['values'],
                      category=DEFAULTS['category'],
                      callback=DEFAULTS['callback'],
                      synopsis=DEFAULTS['synopsis'],
                      environ=DEFAULTS['environ'],
                      registry=DEFAULTS['registry'],
                      mandatory=None,
                      name=DEFAULTS['name'],
                      source=DEFAULTS['source']):
      """
      Declare a command line option

      Instances of subclasses of CommandLineOption must be placed in
      a ConfigManager instance to be used.  See the documentation in
      ConfigManager for more details.

      Keyword Arguments:
      docstring -- a string in the format of Python documentation
         strings that describes the option and its usage.  The first
         line is assumed to be a one-line summary for the option.
         The following paragraphs are assumed to be a complete description
         of the option.  You can give a paragraph with the label
         'Valid Values:' that contains a short description of the
         values that are valid for the current option.  If this
         paragraph exists and an error is encountered while validating
         the option, this paragraph will be printed instead of the
         somewhat generic error message for that option type.

      options -- a string containing all possible variants of the
         option.  All variants should contain the '-', '--', etc. at
         the beginning.  For boolean options, the option can be preceded
         by a '!' to mean that the option should be turned OFF rather
         than ON which is the default.

      default -- value for the option to take if it isn't specified
         on the command line

      optional -- value for the option if it is given without a value.
         This is only used for options that normally take a value,
         but you also want a default that indicates that the option
         was given without a value.

      values -- valid values for the option.  This argument can take
         the following forms:

         single value -- for StringOption this this is a string, for
            IntegerOption this is an integer, for FloatOption this is
            a float.  The single value mode is most useful when the
            value is a regular expression.  For example, to specify
            that a StringOption must be a string of characters followed
            by a digit, 'values' would be set to re.compile(r'\w+\d').

         range of values -- a two element list can be given to specify
            the endpoints of a range of valid values.  This is probably
            most useful on IntegerOption and FloatOption.  For example,
            to specify that an IntegerOption can only take the values
            from 0 to 10, 'values' would be set to [0,10].

            NOTE: This mode must *always* use a Python list since using
                  a tuple means something else entirely.

         tuple of values -- a tuple of values can be used to specify
            a complete list of valid values.  For example, to specify
            that an IntegerOption can take the values 1, 2, or 3, 'values'
            would be set to (1,2,3).  If a string value can only take
            the values, 'hi', 'bye', and any string of characters beginning
            with the letter 'z', 'values' would be set to
            ('hi','bye',re.compile(r'z.*?')).

            NOTE: This mode must *always* use a Python tuple since using
                  a list means something else entirely.

      category -- a category key which specifies which category the
         option belongs to (see Option.add_category())

      callback -- a function to call after the value of the option has
         been validated.  This function will be called with the validated
         option value as its only argument.

      environ -- environment variable to use as default value instead
         of specified value.  If the environment variable exists, it
         will be used for the default value instead of the specified value.

      registry -- registry key to use as default value instead of
         specified value.  If the registry key exists, it will be used
         for the default value instead of the specified value.  A
         specified environment variable takes precedence over this value.

      name -- key used to get the option from its corresponding section.
         You do not need to specify this.  It will be set automatically
         when you put the option into the ConfigManager instance.

      mandatory -- flag used to determine if the option itself is
         required to be present.  The idea of a "mandatory option" is
         a little strange, but I have seen it done.

      source -- flag used to determine whether the option was set
         directly in the ConfigManager instance through Python,
         by a configuration file/command line option, etc.  You do not need
         to specify this, it will be set automatically during parsing.
         This flag should have the value of BUILTIN, COMMANDLINE,
         CONFIGFILE, ENVIRONMENT, REGISTRY, or CODE.


      Examples::
         BooleanOption(
            ''' Display help message ''',
            options = '--help -h',
            callback = usage,  # usage() function must exist prior to this
         )

         BooleanOption(
            ''' Set verbosity ''',
            options = '-v --verbose !-q !--quiet',
         )

         StringOption(
            '''
            IP address option

            This option accepts an IP address to connect to.

            Valid Values:
            '#.#.#.#' where # is a number from 1 to 255

            ''',
            options = '--ip-address',
            values = re.compile(r'\d{1,3}(\.\d{1,3}){3}'),
            default = '127.0.0.0',
            synopsis = '#.#.#.#',
            category = 'network',  # Assumes 'network' category exists
         )

         IntegerOption(
            '''
            Number of seconds to wait before timing out

            Valid Values:
            positive integer

            ''',
            options = '--timeout -t',
            default = 300,
            values = [0,1e9],
            category = 'network',
         )

         IntegerOption(
            '''
            Number of tries to connect to the host before giving up

            Valid Values:
            accepts 1, 2, or 3 retries

            ''',
            options = '--tries',
            default = 1,
            values = (1,2,3),
            category = 'network',
         )

         StringOption(
            '''
            Nonsense option for example purposes only

            Valid Values:
            accepts 'hi', 'bye', or any string beginning with the letter 'z'

            ''',
            options = '--nonsense -n',
            default = 'hi',
            values = ('hi', 'bye', re.compile(r'z.*?')),
         )

      """
      if type(self.__class__) is GenericOption:
         raise ValueError('GenericOption cannot be instantiated ' + \
                           'directly, you must use a subclass.')
      self.parent = None
      self.initialize(locals())
      self.data = None

   def initialize(self, vars):
      """ Initialize all instance variables """
      docstring = vars.get('docstring', DEFAULTS['docstring'])
      options = vars.get('options', DEFAULTS['options'])
      default = vars.get('default', DEFAULTS['default'])
      optional = vars.get('optional', DEFAULTS['optional'])
      values = vars.get('values', DEFAULTS['values'])
      category = vars.get('category', DEFAULTS['category'])
      name = vars.get('name', DEFAULTS['name'])
      source = vars.get('source', DEFAULTS['source'])
      callback = vars.get('callback', DEFAULTS['callback'])
      synopsis = vars.get('synopsis', DEFAULTS['synopsis'])
      environ = vars.get('environ', DEFAULTS['environ'])
      mandatory = vars.get('mandatory')

      if mandatory is None:
         mandatory = isinstance(self, GenericArgument)
      registry = vars.get('registry', DEFAULTS['registry'])

      self.actual = None  # Actual option used on the command-line
      self.name = name
      self.options = options
      if category is None:
         self.category = []
      elif type(category) in [type(list), type(tuple)]:
         self.category = list(category)
      else:
         self.category = [category]
      self.values = values
      if synopsis is not None:
         self.synopsis = synopsis
      self.environ = environ
      self.registry = registry
      self.optional = self.cast(optional)
      self.callback = callback
      self.mandatory = mandatory
      self.source = source   # Flag to indicate where the option came from
      self.occurrences = 0 # Number of times the command-line option was used

      self.summary, self.description, self.error = \
                    self._splitDocString(docstring)

      self.file = None # Variable to hold filename of config file where
                       # the option came from.  This should only be
                       # set if self.source == CONFIGFILE.

      if source <= ENVIRONMENT:
         if self.environ and str(self.environ) in os.environ:
            self.default = self.cast(os.environ[str(self.environ)])
            self.source = ENVIRONMENT
         else:
            self.default = self.cast(default)
      else:
         self.default = None

      self.setValue(self.default)

   def defaultValue(self):
      return self.default

   def getValue(self, value=None):
      """ Return value for option """
      # If specified on the commond-line, always return it.
      if self.data is not None and self.source & COMMANDLINE:
         return self.data

      # See if it was set in the environment
      if self.environ and str(self.environ) in os.environ:
         self.source = ENVIRONMENT
         self.file = None
         return self.cast(os.environ[str(self.environ)])

      # If the option was populated in any other way, return it.
      if self.data is not None:
         return self.data

      # Look for a default
      if self.default is not None:
         self.source = BUILTIN
         self.file = None
         return self.default

      if value is not None:
         self.source = CODE
         self.file = None

      return value

   def matches(self, opt):
      """ Return boolean indicating a match to the given option """
      possibilities = [x.replace('!','') for x in self.getPossibleOptions()]
      for i in possibilities:
         if i.startswith(opt):
            return i
      return 0

   def copy(self):
      """ Make a deep copy of self """
      newcopy = self.__class__()
      for key, value in list(vars(self).copy().items()):
         if key == 'data':
            setattr(newcopy, key, copy.copy(value))
         else:
            setattr(newcopy, key, value)
      return newcopy

   def _splitDocString(self, doc):
      """
      Split documentation string into summary, description, and error

      Required Arguments:
      doc -- a string resembling a Python documentation string

      Returns:
      tuple -- a three element tuple containing the summary, description,
         and error message parsed from the documentation string.

      """
      summary = None
      description = None
      error = None

      if doc.strip():
         parts = doc.strip().split('\n')
         if len(parts) > 2:
            # If there is a blank line after the first line,
            # there is a summary.
            if not(parts[1].strip()):
               summary = parts[0].strip()
               description = '\n'.join(parts[2:])
            else:
               summary = None
               description = doc

         elif len(parts) == 2:
            summary = None
            description = doc

         else:
            summary = doc
            description = None

      if summary: summary = summary.strip()

      if description:
         # Look for error message
         m = re.compile(r'^\s*Valid\s+Values\s*:+\s*(.+)(?:\n\s*\n|\s*$)',
                        re.I|re.S|re.MULTILINE).search(description)
         if m: error = m.group(1).strip()

         # Remove extra space from the front of each line
         spaces = re.compile(r'^\n*(\s*)', re.MULTILINE).findall(description)
         spaces = len(min(spaces))
         lines = description.split('\n')
         for i in range(len(lines)):
            if lines[i].strip():
               lines[i] = lines[i][spaces:]
            else:
               lines[i] = ''
         description = '\n'.join(lines).strip()

      if description and summary:
         description = '%s\n\n%s' % (summary, description)
      elif summary:
         description = summary

      return summary, description, error

   def setParent(self, parent):
       """ Set the parent section """
       self.parent = parent

   def names(self):
      """
      Build a dictionary of common variables needed in message strings

      Returns:
      dictionary -- with common variables populated (see return value)

      """
      return {}
      return {'name': self.name, 'default': self.default,
              'option': self.actual, 'synopsis': self.synopsis}

   def __lt__(self, other): return self.compare(other) < 0
   def __gt__(self, other): return self.compare(other) > 0
   def __le__(self, other): return self.compare(other) <= 0
   def __ge__(self, other): return self.compare(other) >= 0
   def __eq__(self, other): return self.compare(other) == 0
   

   def compare(self, other):
      """ Compare option to another using the specified sort order """
      for attr in self.sort_order:
          sattr = getattr(self, attr)
          oattr = getattr(other, attr)

          if sattr == None: sattr = ''
          if oattr == None: oattr = ''

          if type(sattr) is str and type(oattr) is str:
              sattr = re.sub(r'\W', '', sattr)
              oattr = re.sub(r'\W', '', oattr)

          if sattr < oattr:
              return -1
          elif sattr > oattr:
              return 1
      return 0

   def clearValue(self):
      """ Reset the option value to None """
      self.data = None

   def setValue(self, value):
      """
      Set the value of the option to the given value

      Once the value is set to the new value and validated, the
      specified callback function is called with that value as its
      only argument.

      """
      if value is None:
         self.clearValue()
      else:
         if isinstance(self.callback, collections.Callable):
            value = self.callback(self.cast(value))
         self.data = self.validate(value)

   def __str__(self):
      """ Return string representation of the current option value """
      value = self.getValue()
      if value is not None:
         return str(value)
      return ''

   def __repr__(self):
      """
      Build a string comparable to what may have been given on the command line

      """
      value = self.getValue()
      if value is not None:
         option = self.actual
         if not option:
            args = self.getPossibleOptions()
            if args: option = args[0]
         if option:
            return str('%s %s' % (option, value)).strip()
      return ''

   def acceptsArgument(self):
      """ Return a boolean indicating if the option accepts an argument """
      return 1

   def requiresArgument(self):
      """ Return a boolean indicating if the option requires an argument """
      if self.optional is None:
         return 1
      else:
         return 0

   def getPossibleOptions(self):
      """ Return a list of all possible command line options """
      if self.options:
         return [x.strip() for x in re.sub(r'!\s+', r'!', self.options).split()
                           if x.replace('!','').strip()]
      return []

   def validate(self, value):
      """ Validate the value of the option """
      return self.checkValues(self.cast(value))

   def cast(self, value):
      return value

   def checkValues(self, value):
      """ Check the value against the possible valid values """
      # All values are valid
      if self.values is None: return value

      name = self.name
      if self.actual: name = self.actual

      # Check to see if the value is within the valid range
      if isinstance(self.values, list):
         if not self.values:
            pass
         elif value >= self.values[0] and value <= self.values[-1]:
            return value

         if self.error:
            raise InvalidOptionError(name, value,
                                     msg=self.error % self.names())
         else:
            raise InvalidOptionError(name, value,
                  msg='Given value is not within the valid range (%s, %s)' %
                      (self.values[0], self.values[-1]))

      # Check to see if the value is within the list of valid values
      elif isinstance(self.values, tuple):
         # Look for literal values
         for option in [x for x in self.values
                                if not(isinstance(x, RegexType))]:
            if isinstance(option, str) and option.lower() == value.lower():
               return value
            elif option == value:
               return value

         # Look for regular expressions in the list
         for regex in [x for x in self.values if isinstance(x, RegexType)]:
            if regex.search(value):
               return value

         if self.error:
            raise InvalidOptionError(name, value,
                                     msg=self.error % self.names())
         else:
            raise InvalidOptionError(name, value,
                  msg='Given value is not a valid value. Expecting (%s)' % \
                      (', '.join(map(str, self.values))))

      # Check to see if the value is equal to the only valid value
      elif isinstance(self.values, str) or isinstance(self.values, int) \
           or isinstance(self.values, float):
         if value == self.values:
            return value

         if self.error:
            raise InvalidOptionError(name, value,
                                     msg=self.error % self.names())
         else:
            raise InvalidOptionError(name, value,
                  msg='Given value is not a valid value. Expecting (%s)' % \
                       self.values)

      # Check to see if the value is valid using a regex
      elif isinstance(self.values, RegexType):
         if not isinstance(value, str):
            pass
         elif self.values.search(value):
            return value

         if self.error:
            raise InvalidOptionError(name, value,
                                     msg=self.error % self.names())
         else:
            raise InvalidOptionError(name, value,
                                     msg='Given value is not a valid value')

      # Use the user defined validation function
      elif isinstance(self.values, collections.Callable):
         if self.values(value):
            return value

         if self.error:
            raise InvalidOptionError(name, value,
                                     msg=self.error % self.names())
         else:
            raise InvalidOptionError(name, value,
                                     msg='Given value is not a valid value')

      # Bail out if we don't know what this is
      raise ValueError('Unknown valid values type')


class GenericArgument(GenericOption):
   """
   Base class for command-line arguments

   All command line arguments must subclass from this one.  In addition,
   this is an abstract class.  It cannot be instantiated directly.

   """
#  def __init__(self, docstring=DEFAULTS['docstring'],
#                     default=DEFAULTS['default'],
#                     optional=DEFAULTS['optional'],
#                     values=DEFAULTS['values'],
#                     callback=DEFAULTS['callback'],
#                     synopsis=DEFAULTS['synopsis'],
#                     environ=DEFAULTS['environ'],
#                     registry=DEFAULTS['registry'],
#                     name=DEFAULTS['name'],
#                     source=DEFAULTS['source']):
#     if type(self.__class__) is GenericArgument:
#        raise ValueError, 'GenericArgument cannot be instantiated ' + \
#                          'directly, you must use a subclass.'
#     self.parent = None
#     self.initialize(locals())
#     self.data = None

   def __repr__(self):
      """
      Build a string comparable to what may have been given on the command line

      """
      return str(self)



#!/usr/bin/env python

import os
from plasTeX.ConfigManager.Generic import *
from plasTeX.ConfigManager import TooManyValues
from plasTeX.ConfigManager.String import StringOption
from collections import UserList
import collections


class MultiParser(GenericParser):

   def getArgument(self, args, range=None, delim=None, forcedarg=False):
      if range is None:
         range = self.range[:]
      if delim is None:
         delim = self.delim
      new_args, args = GenericParser.getArgument(self, args, range, delim, forcedarg=forcedarg)
      if type(new_args) in [list, tuple]:
         return new_args, args
      elif new_args is None:
         return [], args
      return [new_args], args


class MultiOption(MultiParser, GenericOption, UserList):
   """
   Multiple configuration option

   Multi options are options delimited by a specified character.  They
   can also be represented by an option specified multiple times.
   All other options, when specified more than once, will overwrite
   their previous value.  Multi options will append values each
   time an option is specified.

   """

   synopsis = 'val1,val2,...'

   def __init__(self, docstring=DEFAULTS['docstring'],
                      options=DEFAULTS['options'],
                      default=None,
                      optional=DEFAULTS['optional'],
                      values=DEFAULTS['values'],
                      category=DEFAULTS['category'],
                      callback=DEFAULTS['callback'],
                      synopsis=DEFAULTS['synopsis'],
                      environ=DEFAULTS['environ'],
                      registry=DEFAULTS['registry'],
                      delim=',',
                      range=None,
                      mandatory=None,
                      name=DEFAULTS['name'],
                      source=DEFAULTS['source'],
                      template=StringOption):
      """
      Initialize a multi option

      This class is initialized with the same options as the
      Option class with one addition: delim.  The 'delim' argument
      specifies what the delimiter is for each item in the list.
      If the delimiter is 'None' or whitespace, each item in the
      list is assumed to be delimited by whitespace.

      """
      self.delim = delim
      default = default = []
      self.range = range or [1, '*']
      assert not issubclass(template, MultiOption), \
             'MultiOptions can not have a MultiOption as a template'
      assert issubclass(template, GenericOption), \
             'Templates must be a subclass of GenericOption'
      self.template = template(options=options,name=name,values=values)
      UserList.__init__(self, [])
      GenericOption.initialize(self, locals())

   def cast(self, arg):
      if arg is None: return []
      if type(arg) in [list,tuple]:
         return [self.template.cast(x) for x in list(arg)]
      delim = self.delim
      if not delim:
         delim = ' '
      return [self.template.cast(v.strip()) for v in str(arg).split(delim)
                                            if v.strip()]

   def getValue(self, value=None):
      """ Return value for option """
      if self.data and self.source & COMMANDLINE:
         return self.data

      if self.environ and str(self.environ) in os.environ:
         self.source = ENVIRONMENT
         self.file = None
         return self.cast(os.environ[str(self.environ)])

      if self.data:
         return self.data

      if self.default:
         self.source = BUILTIN
         self.file = None
         return self.default

      self.source = CODE
      self.file = None

      if value is None:
         return []

      return value

   def clearValue(self):
      """ Clear the value to be unset """
      self.data = []

   def __repr__(self):
      """ Print command-line representation """
      delim = self.delim
      if not delim:
         delim = ' '
      if self.data:
         option = self.actual
         if not option:
            args = self.getPossibleOptions()
            if args: option = args[0]
         if option:
            return str('%s %s' % (option, delim.join(self.data))).strip()
      return ''

   def __iadd__(self, other):
      """ Append a value to the list """
      if isinstance(self.callback, collections.Callable):
         other = self.callback(self.cast(other))
      self.data += self.validate(other)
      range = self.validateRange(self.range)
      name = self.name
      if self.actual: name = self.actual
      if len(self.data) > range[1]:
         raise TooManyValues("Expecting at most %s values for option '%s'." % (range[1], name))
      return self

   def validate(self, arg):
      """ Validate the value of the option """
      new_values = []
      for i in self.cast(arg):
#        new_values.append(self.checkValues(i))
         new_values.append(self.template.validate(i))
      return new_values

   def checkValues(self, value):
      return self.template.checkValues(value)

   def __str__(self):
      if self.delim and self.delim.strip():
         delim = '%s ' % self.delim
         return delim.join([str(x) for x in self.data])
      else:
         return '\n'.join([str(x) for x in self.data])
      return str(self.data)

   def acceptsArgument(self):
      """ Return a boolean indicating if the option accepts an argument """
      range = self.validateRange(self.range)
      return not(not(range[1]))

   def requiresArgument(self):
      """ Return a boolean indicating if the option requires an argument """
      range = self.validateRange(self.range)
      return not(not(range[0]))

   def setValue(self, value):
      """
      Set the value of the option to the given value

      Once the value is set to the new value and validated, the
      specified callback function is called with that value as its
      only argument.

      """
      if value is None or ((type(value) in [list,tuple]) and not(value)):
         self.clearValue()
      else:
         if isinstance(self.callback, collections.Callable):
            value = self.callback(self.cast(value))
         self.data = self.validate(value)


class MultiArgument(GenericArgument, MultiOption):
   """ Multiple command-line option """

   def __init__(self, docstring=DEFAULTS['docstring'],
                      options=DEFAULTS['options'],
                      default=None,
                      optional=DEFAULTS['optional'],
                      values=DEFAULTS['values'],
                      category=DEFAULTS['category'],
                      callback=DEFAULTS['callback'],
                      synopsis=DEFAULTS['synopsis'],
                      environ=DEFAULTS['environ'],
                      registry=DEFAULTS['registry'],
                      delim=' ',
                      range=None,
                      mandatory=None,
                      name=DEFAULTS['name'],
                      source=DEFAULTS['source'],
                      template=StringOption):
      """ Initialize a multi argument """
      self.delim = delim
      default = default or []
      self.range = range or [1, '*']
      assert not issubclass(template, MultiArgument), \
             'MultiArguments can not have a MultiArguments as a template'
      assert not issubclass(template, MultiOption), \
             'MultiOptions can not have a MultiOptions as a template'
      assert issubclass(template, GenericOption), \
             'Templates must be a subclass of GenericOption'
      self.template = template(options=options,name=name,values=values)
      UserList.__init__(self, [])
      GenericOption.initialize(self, locals())



#!/usr/bin/env python

from plasTeX.Tokenizer import Token
from plasTeX import Command, Environment


class alltt(Environment):

    def invoke(self, tex):
        self.ownerDocument.context.push(self)
        esc = self.ownerDocument.context.categories[Token.CC_ESCAPE]
        begin = self.ownerDocument.context.categories[Token.CC_BGROUP]
        end = self.ownerDocument.context.categories[Token.CC_EGROUP]
        self.ownerDocument.context.setVerbatimCatcodes()
        for i in esc:
            self.ownerDocument.context.catcode(i, Token.CC_ESCAPE)
        for i in begin:
            self.ownerDocument.context.catcode(i, Token.CC_BGROUP)
        for i in end:
            self.ownerDocument.context.catcode(i, Token.CC_EGROUP)
        Environment.invoke(self, tex)
        self.ownerDocument.context.pop(self)




#!/usr/bin/env python

import os, time, tempfile, shutil, re, string, pickle
try: from hashlib import md5
except ImportError: from md5 import new as md5
from plasTeX.Logging import getLogger
from io import StringIO
from plasTeX.Filenames import Filenames
from collections import OrderedDict
import subprocess
import shlex

log = getLogger()
depthlog = getLogger('render.images.depth')
status = getLogger('status')
imagelog = getLogger('imager')

try:
    from PIL import Image as PILImage
    from PIL import ImageChops as PILImageChops
except ImportError:
    PILImage = PILImageChops = None

def autoCrop(im, bgcolor=None, margin=0):
    """
    Automatically crop image down to non-background portion

    Required Argument:
    im -- image object

    Optional Argument:
    bgcolor -- value or tuple containing the color to use for the
        background color when cropping
    margin -- leave this many pixels around the content.  If there
        aren't that many pixels to leave, leave as many as possible.

    Returns: cropped image object and tuple containing the number
        of pixels removed from each side (left, top, right, bottom)

    """
    if im.mode != "RGB":
        im = im.convert("RGB")

    origbbox = im.getbbox()
    if origbbox is None:
        origbbox = (0,0,im.size[0],im.size[1])

    # Figure out the background color from the corners, if needed
    if bgcolor is None:
        topleft = im.getpixel((origbbox[0],origbbox[1]))
        topright = im.getpixel((origbbox[2]-1,origbbox[1]))
        bottomleft = im.getpixel((origbbox[0],origbbox[3]-1))
        bottomright = im.getpixel((origbbox[2]-1,origbbox[3]-1))
        corners = [topleft, topright, bottomleft, bottomright]

        matches = []
        matches.append(len([x for x in corners if x == topleft]))
        matches.append(len([x for x in corners if x == topright]))
        matches.append(len([x for x in corners if x == bottomleft]))
        matches.append(len([x for x in corners if x == bottomright]))

        try: bgcolor = corners[matches.index(1)]
        except ValueError: pass
        try: bgcolor = corners[matches.index(2)]
        except ValueError: pass
        try: bgcolor = corners[matches.index(3)]
        except ValueError: pass
        try: bgcolor = corners[matches.index(4)]
        except ValueError: pass

    # Create image with only the background color
    bg = PILImage.new("RGB", im.size, bgcolor)

    # Get bounding box of non-background content
    diff = PILImageChops.difference(im, bg)
    bbox = diff.getbbox()
    if bbox:
        if margin:
            bbox = list(bbox)
            bbox[0] -= margin
            bbox[1] -= margin
            bbox[2] += margin
            bbox[3] += margin
            bbox = tuple([max(0,x) for x in bbox])
        return im.crop(bbox), tuple([abs(x-y) for x,y in zip(origbbox,bbox)]), bgcolor
    return PILImage.new("RGB", (1,1), bgcolor), (0,0,0,0), bgcolor
    return None, None, bgcolor # no contents

class Box(object):
    pass

class Dimension(float):
    """
    Dimension object used for width, height, and depth of images

    This object is simply a float value.  The value of the float
    is in pixels.  All other units can be gotten using their
    corresponding property.

    """
    fontSize = 15
    @property
    def ex(self):
        return '%sex' % self.format(self / (self.fontSize * 0.6))
    @property
    def em(self):
        return '%sem' % self.format(self / self.fontSize)
    @property
    def pt(self):
        return '%spt' % self.format(self)
    @property
    def px(self):
        return '%spx' % self.format(self)
    @property
    def mm(self):
        return '%smm' % self.format(self.cm * 10.0)
    @property
    def inch(self):
        return '%sin' % self.format(self.pt / 72.0)
    @property
    def cm(self):
        return '%scm' % self.format(self.inch * 2.54)
    @property
    def pc(self):
        return '%spc' % self.format(self.pt / 12.0)

    def __getattribute__(self, name):
        if name in ['in']:
            return self.inch
        return float.__getattribute__(self, name)

    def format(self, value):
        if abs(int(value) - value) < 0.0001:
            return '%s' % int(value)
        return '%0.3f' % value

    def __str__(self):
        return self.format(self)

    def __repr__(self):
        return self.format(self)


class DimensionPlaceholder(str):
    """
    Placeholder for dimensions

    Dimensions for an image aren't generally known until the end of
    the rendering process.  This object generates a placeholder
    for the dimension.

    """
    imageUnits = ''
    def __getattribute__(self, name):
        if name in ['in','ex','em','pt','px','mm','cm','pc']:
            if not self:
                return self
            vars = {'units':name}
            return self + string.Template(self.imageUnits).substitute(vars)
        return str.__getattribute__(self, name)
    def __setattribute__(self, name, value):
        if name in ['in','ex','em','pt','px','mm','cm','pc']:
            return
        return str.__setattribute__(self, name, value)

class Image(object):
    """ Generic image object """

    def __init__(self, filename, config, width=None, height=None, alt=None,
                       depth=None, longdesc=None):
        self.filename = filename
        self.path = os.path.join(os.getcwd(), self.filename)
        self.width = width
        self.height = height
        self.alt = alt
        self.depth = depth
        self.depthRatio = 0
        self.longdesc = longdesc
        self.config = config
        self._cropped = False
        self.bitmap = self
        self.checksum = None

    def height():
        def fget(self):
            return getattr(self.bitmap, '_height', None)
        def fset(self, value):
            if value is None:
                self._height = value
            elif isinstance(value, DimensionPlaceholder):
                self._height = value
            else:
                self._height = Dimension(value)
        return locals()
    height = property(**height())

    def width():
        def fget(self):
            return getattr(self.bitmap, '_width', None)
        def fset(self, value):
            if value is None:
                self._width = value
            elif isinstance(value, DimensionPlaceholder):
                self._width = value
            else:
                self._width = Dimension(value)
        return locals()
    width = property(**width())

    def depth():
        def fget(self):
            return getattr(self, '_depth', None)
        def fset(self, value):
            if value is None:
                self._depth = value
            elif isinstance(value, DimensionPlaceholder):
                self._depth = value
            else:
                self._depth = Dimension(value)
        return locals()
    depth = property(**depth())

    @property
    def url(self):
        base = self.config['base-url']
        if base and base.endswith('/'):
            base = base[:-1]
        if base:
            return '%s/%s' % (base, self.filename)
        return self.filename

    def crop(self):
        """ Do the actual cropping """
        if self._cropped:
            return

        # Crop an SVG image
        if os.path.splitext(self.path)[-1] in ['.svg']:
            svg = open(self.path,'r').read()

            self.width = 0
            width = re.search(r'width=(?:\'|")([^\d\.]+)\w*(?:\'|")', svg)
            if width:
                self.width = float(width)

            self.height = 0
            height = re.search(r'height=(?:\'|")([^\d\.]+)\w*(?:\'|")', svg)
            if height:
                self.height = float(height)

            self.depth = 0
            if self.bitmap and self.height:
                depth = (self.height / self.bitmap.height) * self.bitmap.depth
                if abs(depth - int(depth)) > 0.1:
                    self.depth = depth - 1
                else:
                    self.depth = depth

            self._cropped = True
            return

        padbaseline = self.config['baseline-padding']

        try:
            im, self.depth = self._stripBaseline(PILImage.open(self.path),
                                             padbaseline)
            self.width, self.height = im.size
        except IOError as msg:
#           import traceback
#           traceback.print_exc()
            self._cropped = True
            log.warning(msg)
            return

        if padbaseline and self.depth > padbaseline:
            log.warning('depth of image %s (%d) is greater than the baseline padding (%s).  This may cause the image to be misaligned with surrounding text.', self.filename, self.depth, padbaseline)

        if self.config['transparent']:
            im = im.convert("P")
            lut = im.resize((256,1))
            lut.putdata(list(range(256)))
            index = list(lut.convert("RGB").getdata()).index((255,255,255))
            im.save(self.path, transparency=index)
        else:
            im.save(self.path)

        self._cropped = True

    def __str__(self):
        return self.filename

    def __repr__(self):
        return self.filename

    def _autoCrop(self, im, bgcolor=None, margin=0):
        return autoCrop(im, bgcolor, margin)

    def _stripBaseline(self, im, padbaseline=0):
        """
        Find the baseline register mark and crop it out

        The image has to have a particular layout.  The top left corner
        must be the background color of the image.  There should be a
        square registration mark which has the bottom edge at the baseline
        of the image (see \\plasTeXregister in LaTeX code at the top
        of this file).  This registration mark should be the leftmost
        content of the image.  If the registration mark is at the top
        of the image, the baseline is ignored.

        Required Arguments:
        im -- image to be cropped

        Keyword Arguments:
        padbaseline -- amount to pad the bottom of all cropped images.
            This allows you to use one margin-bottom for all images;
            however, you need to make sure that this padding is large
            enough to handle the largest descender in the document.

        Returns:
        (cropped image, distance from baseline to bottom of image)

        """
        if im.mode != "RGB":
            im = im.convert("RGB")

        depth = 0

        # Crop the image so that the regitration mark is on the left edge
        im, box, background = self._autoCrop(im)

        width, height = im.size

        # Determine if registration mark is at top or left
        top = False
        # Found mark at top
        if im.getpixel((0,0)) != background:
            top = True
            i = 1
            # Parse past the registration mark
            # We're fudging the vertical position by 1px to catch
            # things sitting right under the baseline.
            while i < width and im.getpixel((i,1)) != background:
                i += 1
            # Look for additional content after mark
            if i < width:
                while i < width and im.getpixel((i,1)) == background:
                    i += 1
                # If there is non-background content after mark,
                # consider the mark to be on the left
                if i < width:
                    top = False

        # Registration mark at the top
        blank = False
        if top:
            pos = height - 1
            while pos and im.getpixel((0,pos)) == background:
                pos -= 1
            depth = pos - height + 1

            # Get the height of the registration mark so it can be cropped out
            rheight = 0
            while rheight < height and im.getpixel((0,rheight)) != background:
                rheight += 1

            # If the depth is the entire height, just make depth = 0
            if -depth == (height-rheight):
                depth = 0

            # Handle empty images
            bbox = im.getbbox()
            if bbox is None or rheight == (height-1):
                blank = True
            else:
                bbox = list(bbox)
                bbox[1] = rheight

        # Registration mark on left side
        if blank or not(top) or im.getbbox()[1] == 0:
            pos = height - 1
            while pos and im.getpixel((0,pos)) == background:
                pos -= 1
            depth = pos - height + 1

            # Get the width of the registration mark so it can be cropped out
            rwidth = 0
            while rwidth < width and im.getpixel((rwidth,pos)) != background:
                rwidth += 1

            # Handle empty images
            bbox = im.getbbox()
            if bbox is None or rwidth == (width-1):
                return PILImage.new("RGB", (1,1), background), 0

            bbox = list(bbox)
            bbox[0] = rwidth

        # Crop out register mark, and autoCrop result
        im, cropped, background = self._autoCrop(im.crop(bbox), background)

        # If the content was entirely above the baseline,
        # we need to keep that whitespace
        depth += cropped[3]
        depthlog.debug('Depth of image %s is %s', self.filename, depth)

        # Pad all images with the given amount.  This allows you to
        # set one margin-bottom for all images.
        if padbaseline:
            width, height = im.size
            newim = PILImage.new("RGB", (width,height+(padbaseline+depth)), background)
            newim.paste(im, im.getbbox())
            im = newim

        return im, depth


class Imager(object):
    """ Generic Imager """

    # The command to run on the LaTeX output file to generate images.
    # This should be overridden by the subclass.
    command = '' # dvipng -o img%d.png -D 120 -Q 4'

    # The compiler command used to compile the LaTeX document
    compiler = 'latex'

    # Verification command to determine if the imager is available
    verification = ''

    fileExtension = '.png'

    imageAttrs = ''
    imageUnits = ''

    def __init__(self, document, imageTypes=None):
        self.config = document.config
        self.ownerDocument = document

        if imageTypes is None:
            self.imageTypes = [self.fileExtension]
        else:
            self.imageTypes = imageTypes[:]

        # Dictionary that makes sure each image is only generated once.
        # The key is the LaTeX source and the value is the image instance.
        self._cache = {}
        usednames = {}
        self._filecache = os.path.abspath(os.path.join('.cache',
                                          self.__class__.__name__+'.images'))
        if self.config['images']['cache'] and os.path.isfile(self._filecache):
            try:
                self._cache = pickle.load(open(self._filecache, 'r'))
                for key, value in list(self._cache.items()):
                    if not os.path.isfile(value.filename):
                        del self._cache[key]
                        continue
                    usednames[value.filename] = None
            except ImportError:
                os.remove(self._filecache)

        # List of images in the order that they appear in the LaTeX file
        self.images = OrderedDict()

        # Images that are simply copied from the source directory
        self.staticimages = OrderedDict()

        # Filename generator
        self.newFilename = Filenames(self.config['images'].get('filenames', raw=True),
                           variables={'jobname':document.userdata.get('jobname','')},
                           extension=self.fileExtension, invalid=usednames)

        # Start the document with a preamble
        self.source = StringIO()
        self.source.write('\\scrollmode\n')
        self.writePreamble(document)
        self.source.write('\\begin{document}\n')

        # Set up additional options
        self._configOptions = self.formatConfigOptions(self.config['images'])

    def formatConfigOptions(self, config):
        """
        Format configuration options as command line options

        Required Arguments:
        config -- the images section of the configuration object

        Returns: a list of two-element tuples contain option value pairs

        Example::
            output = []
            if config['resolution']:
                output.append(('-D', config['resolution']))
            return output

        """
        return []

    def writePreamble(self, document):
        """ Write any necessary code to the preamble of the document """
        self.source.write(document.preamble.source)
        self.source.write('\\makeatletter\\oddsidemargin -0.25in\\evensidemargin -0.25in\n')

#       self.source.write('\\tracingoutput=1\n')
#       self.source.write('\\tracingonline=1\n')
#       self.source.write('\\showboxbreadth=\maxdimen\n')
#       self.source.write('\\showboxdepth=\maxdimen\n')
#       self.source.write('\\newenvironment{plasTeXimage}[1]{\\def\\@current@file{#1}\\thispagestyle{empty}\\def\\@eqnnum{}\\setbox0=\\vbox\\bgroup}{\\egroup\\typeout{imagebox:\\@current@file(\\the\\ht0+\\the\\dp0)}\\box0\\newpage}')

        self.source.write('\\@ifundefined{plasTeXimage}{'
                          '\\newenvironment{plasTeXimage}[1]{' +
                          '\\vfil\\break\\plasTeXregister' +
                          '\\thispagestyle{empty}\\def\\@eqnnum{}\\def\\tagform@{\\@gobble}' +
                          '\\ignorespaces}{}}{}\n')
        self.source.write('\\@ifundefined{plasTeXregister}{' +
                          '\\def\\plasTeXregister{\\parindent=-0.5in\\ifhmode\\hrule' +
                          '\\else\\vrule\\fi height 2pt depth 0pt ' +
                          'width 2pt\\hskip2pt}}{}\n')

    def verify(self):
        """ Verify that this commmand works on this machine """
        if self.verification:
            proc = os.popen(self.verification)
            proc.read()
            if not proc.close():
                return True
            return False

        if not self.command.strip():
            return False

        cmd = self.command.split()[0]

        if not cmd:
            return False

        proc = os.popen('%s --help' % cmd)
        proc.read()
        if not proc.close():
            return True

        return False

    @property
    def enabled(self):
        if self.config['images']['enabled'] and \
           (self.command or (type(self) is not Imager and type(self) is not VectorImager)):
            return True
        return False

    def close(self):
        """ Invoke the rendering code """
        # Finish the document
        self.source.write('\n\\end{document}\\endinput')

        for value in list(self._cache.values()):
            if value.checksum and os.path.isfile(value.path):
                 d = md5(open(value.path,'r').read()).digest()
                 if value.checksum != d:
                     log.warning('The image data for "%s" on the disk has changed.  You may want to clear the image cache.' % value.filename)
        # Bail out if there are no images
        if not self.images:
            return

        if not self.enabled:
            return

        # Compile LaTeX source, then convert the output
        self.source.seek(0)
        output = self.compileLatex(self.source.read())
        if output is None:
            log.error('Compilation of the document containing the images failed.  No output file was found.')
            return

        self.convert(output)

        for value in list(self._cache.values()):
            if value.checksum is None and os.path.isfile(value.path):
                 value.checksum = md5(open(value.path,'rb').read()).digest()

        if not os.path.isdir(os.path.dirname(self._filecache)):
            os.makedirs(os.path.dirname(self._filecache))
        pickle.dump(self._cache, open(self._filecache,'wb'))

    def compileLatex(self, source):
        """
        Compile the LaTeX source

        Arguments:
        source -- the LaTeX source to compile

        Returns:
        file object corresponding to the output from LaTeX

        """
        cwd = os.getcwd()
        # Make a temporary directory to work in
        tempdir = tempfile.mkdtemp()
        os.chdir(tempdir)

        filename = 'images.tex'
        # Write LaTeX source file
        encoding = self.config['files']['input-encoding']
        self.source.seek(0)
        with open(filename, 'w', encoding=encoding) as f:
            f.write(self.source.read())

        # Run LaTeX
        os.environ['SHELL'] = '/bin/sh'
        program = self.config['images']['compiler']
        if not program:
            program = self.compiler

        cmd = r'%s %s' % (program, filename)
        p = subprocess.Popen(shlex.split(cmd),
                     stdout=subprocess.PIPE,
                     stderr=subprocess.STDOUT,
                     universal_newlines=True
                     )
        while True:
            line = p.stdout.readline()
            done = p.poll()
            if line:
                imagelog.info(line.strip())
            elif done is not None:
                break

        output = None
        for ext in ['.dvi','.pdf','.ps']:
            if os.path.isfile('images'+ext):
                output = open('images'+ext, 'rb')
                break

        # Change back to original working directory
        os.chdir(cwd)

        return output

    def executeConverter(self, output):
        """
        Execute the actual image converter

        Arguments:
        output -- file object pointing to the rendered LaTeX output

        Returns:
        two-element tuple.  The first element is the return code of the
        command.  The second element is the list of filenames generated.
        If the default filenames (i.e. img001.png, img002.png, ...) are
        used, you can simply return None.

        """
        open('images.out', 'wb').write(output.read())
        options = ''
        if self._configOptions:
            for opt, value in self._configOptions:
                opt, value = str(opt), str(value)
                if ' ' in value:
                    value = '"%s"' % value
                options += '%s %s ' % (opt, value)

        cmd = r'%s %s%s' % (self.command, options, 'images.out')
        p = subprocess.Popen(shlex.split(cmd),
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             universal_newlines=True
                           )
        done = None
        while True:
            line = p.stdout.readline()
            done = p.poll()
            if line:
                imagelog.info(str(line.strip()))
            elif done is not None:
                break
        return done, None

    def convert(self, output):
        """
        Convert the output from LaTeX into images

        Arguments:
        output -- output file object

        """
        if not self.command and self.executeConverter is Imager.executeConverter:
            log.warning('No imager command is configured.  ' +
                        'No images will be created.')
            return

        cwd = os.getcwd()

        # Make a temporary directory to work in
        tempdir = tempfile.mkdtemp()
        os.chdir(tempdir)

        # Execute converter
        rc, images = self.executeConverter(output)
        if rc:
            log.warning('Image converter did not exit properly.  ' +
                        'Images may be corrupted or missing.')

        # Get a list of all of the image files
        if images is None:
            images = [f for f in os.listdir('.')
                            if re.match(r'^img\d+\.\w+$', f)]
        if len(images) != len(self.images):
            log.warning('The number of images generated (%d) and the number of images requested (%d) is not the same.' % (len(images), len(self.images)))

        # Sort by creation date
        #images.sort(lambda a,b: cmp(os.stat(a)[9], os.stat(b)[9]))

        images.sort(key=lambda x: int(re.search(r'(\d+)\.\w+$',x).group(1)))
        os.chdir(cwd)

        if PILImage is None:
            log.warning('PIL (Python Imaging Library) is not installed.  ' +
                        'Images will not be cropped.')

        # Move images to their final location
        for src, dest in zip(images, list(self.images.values())):
            # Move the image
            directory = os.path.dirname(dest.path)
            if directory and not os.path.isdir(directory):
                os.makedirs(directory)
            try:
                shutil.copy2(os.path.join(tempdir,src), dest.path)
            except OSError:
                shutil.copy(os.path.join(tempdir,src), dest.path)

            # Crop the image
            try:
                dest.crop()
                status.dot()
            except Exception as msg:
                import traceback
                traceback.print_exc()
                log.warning('failed to crop %s (%s)', dest.path, msg)

        # Remove temporary directory
        shutil.rmtree(tempdir, True)

    def writeImage(self, filename, code, context):
        """
        Write LaTeX source for the image

        Arguments:
        filename -- the name of the file that will be generated
        code -- the LaTeX code of the image
        context -- the LaTeX code of the context of the image

        """
        self.source.write('%s\n\\begin{plasTeXimage}{%s}\n%s\n\\end{plasTeXimage}\n' % (context, filename, code))

    def newImage(self, text, context='', filename=None):
        """
        Invoke a new image

        Required Arguments:
        text -- the LaTeX source to be rendered in an image

        Keyword Arguments:
        context -- LaTeX source to be executed before the image
            is created.  This generally consists of setting counters,
            lengths, etc. that will be used by the code that
            generates the image.
        filename -- filename to force the image to.  This filename
            should not include the file extension.

        """
        # Convert ligatures back to original string
        for dest, src in self.ownerDocument.charsubs:
            text = text.replace(src, dest)

        key = text

        # See if this image has been cached
        if key in list(self._cache.keys()):
            return self._cache[key]

        # Generate a filename
        if not filename:
            filename = self.newFilename()

        # Add the image to the current document and cache
        #log.debug('Creating %s from %s', filename, text)
        self.writeImage(filename, text, context)

        img = Image(filename, self.config['images'])

        # Populate image attrs that will be bound later
        if self.imageAttrs:
            tmpl = string.Template(self.imageAttrs)
            vars = {'filename':filename}
            for name in ['height','width','depth']:
                if getattr(img, name) is None:
                    vars['attr'] = name
                    value = DimensionPlaceholder(tmpl.substitute(vars))
                    value.imageUnits = self.imageUnits
                    setattr(img, name, value)

        self.images[filename] = self._cache[key] = img
        return img

    def getImage(self, node):
        """
        Get an image from the given node whatever way possible

        This method attempts to find an existing image using the
        `imageoverride' attribute.  If it finds it, the image is
        converted to the appropriate output format and copied to the
        images directory.  If no image is available, or there was
        a problem in getting the image, an image is generated.

        Arguments:
        node -- the node to create the image from

        Returns:
        Image instance

        """
        name = getattr(node, 'imageoverride', None)
        if name is None:
            return self.newImage(node.source)

        if name in list(self.staticimages.keys()):
            return self.staticimages[name]

        # Copy or convert the image as needed
        path = self.newFilename()
        newext = os.path.splitext(path)[-1]
        oldext = os.path.splitext(name)[-1]
        try:
            directory = os.path.dirname(path)
            if directory and not os.path.isdir(directory):
                os.makedirs(directory)

            # If PIL isn't available or no conversion is necessary,
            # just copy the image to the new location
            if newext == oldext or oldext in self.imageTypes:
                path = os.path.splitext(path)[0] + os.path.splitext(name)[-1]
                if PILImage is None:
                    shutil.copyfile(name, path)
                    tmpl = string.Template(self.imageAttrs)
                    width = DimensionPlaceholder(tmpl.substitute({'filename':path, 'attr':'width'}))
                    height = DimensionPlaceholder(tmpl.substitute({'filename':path, 'attr':'height'}))
                    height.imageUnits = width.imageUnits = self.imageUnits
                else:
                    if os.path.splitext(name)[1].lower() == '.svg':
                        shutil.copyfile(name, path)
                        width = height = None
                    else:
                        img = PILImage.open(name)
                        width, height = img.size
                        scale = self.config['images']['scale-factor']
                        if scale != 1:
                            width = int(width * scale)
                            height = int(height * scale)
                            img.resize((width,height))
                            img.save(path)
                        else:
                            shutil.copyfile(name, path)

            # If PIL is available, convert the image to the appropriate type
            else:
                img = PILImage.open(name)
                width, height = img.size
                scale = self.config['images']['scale-factor']
                if scale != 1:
                    width = int(width * scale)
                    height = int(height * scale)
                    img.resize((width,height))
                img.save(path)
            img = Image(path, self.ownerDocument.config['images'], width=width, height=height)
            self.staticimages[name] = img
            return img

        # If anything fails, just let the imager handle it...
        except Exception as msg:
            #log.warning('%s in image "%s".  Reverting to LaTeX to generate the image.' % (msg, name))
            pass
        return self.newImage(node.source)


class VectorImager(Imager):
    fileExtension = '.svg'

    def writePreamble(self, document):
        Imager.writePreamble(self, document)
#       self.source.write('\\usepackage{type1ec}\n')
        self.source.write('\\def\\plasTeXregister{}\n')



#!/usr/bin/env python

"""
C.8.4 Numbering (p194)

"""

from plasTeX import Command, Environment
from plasTeX.Logging import getLogger

class newcounter(Command):
    args = 'name:str [ within ]'
    def invoke(self, tex):
        a = self.parse(tex)
        self.ownerDocument.context.newcounter(a['name'], a['within'])

class setcounter(Command):
    args = 'name:str value:int'
    def invoke(self, tex):
        a = self.parse(tex)
        self.ownerDocument.context.counters[a['name']].setcounter(a['value'])

class addtocounter(Command):
    args = 'name:str value:int'
    def invoke(self, tex):
        a = self.parse(tex)
        self.ownerDocument.context.counters[a['name']].addtocounter(a['value'])

class value(Command):
    args = 'name:str'
    def invoke(self, tex):
        a = self.parse(tex)
        return tex.textTokens(self.ownerDocument.context.counters[a['name']].arabic)

class arabic(Command):
    """ Return arabic representation """
    args = 'name:str'
    def invoke(self, tex):
        a = self.parse(tex)
        return tex.textTokens(self.ownerDocument.context.counters[a['name']].arabic)

class Roman(Command):
    """ Return uppercase roman representation """
    args = 'name:str'
    def invoke(self, tex):
        a = self.parse(tex)
        roman = ""
        n, number = divmod(int(self.ownerDocument.context.counters['name']), 1000)
        roman = "M"*n
        if number >= 900:
            roman = roman + "CM"
            number = number - 900
        while number >= 500:
            roman = roman + "D"
            number = number - 500
        if number >= 400:
            roman = roman + "CD"
            number = number - 400
        while number >= 100:
            roman = roman + "C"
            number = number - 100
        if number >= 90:
            roman = roman + "XC"
            number = number - 90
        while number >= 50:
            roman = roman + "L"
            number = number - 50
        if number >= 40:
            roman = roman + "XL"
            number = number - 40
        while number >= 10:
            roman = roman + "X"
            number = number - 10
        if number >= 9:
            roman = roman + "IX"
            number = number - 9
        while number >= 5:
            roman = roman + "V"
            number = number - 5
        if number >= 4:
            roman = roman + "IV"
            number = number - 4
        while number > 0:
            roman = roman + "I"
            number = number - 1
        return tex.textTokens(roman)

class roman(Roman):
    """ Return the lowercase roman representation """
    def invoke(self, tex):
        a = self.parse(tex)
        return tex.textTokens(self.ownerDocument.context.counters[a['name']].roman)

class Alph(Command):
    """ Return the uppercase letter representation """
    args = 'name:str'
    def invoke(self, tex):
        a = self.parse(tex)
        return tex.textTokens(self.ownerDocument.context.counters[a['name']].Alph)

class alph(Alph):
    """ Return the lowercase letter representation """
    def invoke(self, tex):
        a = self.parse(tex)
        return tex.textTokens(self.ownerDocument.context.counters[a['name']].alph)

class fnsymbol(Command):
    """ Return the symbol representation """
    args = 'name:str'
    def invoke(self, tex):
        a = self.parse(tex)
        return tex.textTokens(self.ownerDocument.context.counters[a['name']].fnsymbol)

class stepcounter(Command):
    args = 'name:str'
    def invoke(self, tex):
        a = self.parse(tex)
        return self.ownerDocument.context.counters[a['name']].stepcounter()

class refstepcounter(Command):
    args = 'name:str'
    def invoke(self, tex):
        a = self.parse(tex)
        return self.ownerDocument.context.counters[a['name']].stepcounter()



#!/usr/bin/env python

from plasTeX import Command, Environment

class afterpage(Command):
    args = 'self:nox'

    def invoke(self, tex):
        super(afterpage, self).invoke(tex)
        return []



from plasTeX import Base
class Abstract(Base.Command):
    args = 'argument'

class Acceptdate(Base.Command):
    args = 'argument'

class Address(Base.Command):
    args = 'argument'

class Bookauthor(Base.Command):
    args = 'argument'

class Booktitle(Base.Command):
    args = 'argument'

class code(Base.Command):
    args = 'argument'

class Code(Base.verbatim):
    pass

class CodeChunk(Base.Environment):
    pass

class CodeInput(Base.verbatim):
    pass

class CodeOutput(Base.verbatim):
    pass

class COV(Base.Command):
    pass

class doi(Base.Command):
    args = 'argument'

class email(Base.Command):
    args = 'argument'

class E(Base.Command):
    pass

class Firstdate(Base.Command):
    args = 'argument'

class Footername(Base.Command):
    args = 'argument'

class Hyperauthor(Base.Command):
    args = 'argument'

class Hypersubject(Base.Command):
    args = 'argument'

class ISBN(Base.Command):
    args = 'argument'

class Issue(Base.Command):
    args = 'argument'

class Keywords(Base.Command):
    args = 'argument'

class Month(Base.Command):
    args = 'argument'

class Pages(Base.Command):
    args = 'argument'

class pkg(Base.Command):
    args = 'argument'

class Plaintitle(Base.Command):
    args = 'argument'

class Plainauthor(Base.Command):
    args = 'argument'

class Plainkeywords(Base.Command):
    args = 'argument'

class Plainreviewer(Base.Command):
    args = 'argument'

class Price(Base.Command):
    args = 'argument'

class Prob(Base.Command):
    pass

class proglang(Base.Command):
    args = 'argument'

class Pubaddress(Base.Command):
    args = 'argument'

class Publisher(Base.Command):
    args = 'argument'

class Pubyear(Base.Command):
    args = 'argument'

class Reviewauthor(Base.Command):
    args = 'argument'

class Reviewer(Base.Command):
    args = 'argument'

class Schunk(Base.Environment):
    pass

class Scode(Base.verbatim):
    pass

class Seconddate(Base.Command):
    args = 'argument'

class Seriesname(Base.Command):
    args = 'argument'

class Shorttitle(Base.Command):
    args = 'argument'

class Sinput(Base.verbatim):
    pass

class Softwaretitle(Base.Command):
    args = 'argument'

class Soutput(Base.verbatim):
    pass

class Submitdate(Base.Command):
    args = 'argument'

class VAR(Base.Command):
    pass

class Volume(Base.Command):
    args = 'argument'

class Year(Base.Command):
    args = 'argument'





#!/usr/bin/env python

import sys, re
from plasTeX import Base

try: import pygments
except: pygments = None

class listingsname(Base.Command):
    str = 'Listing'

PackageOptions = {}

def ProcessOptions(options, document):
    document.context.newcounter('listings',
                                resetby='chapter',
                                format='${thechapter}.${listings}')
    PackageOptions.update(options)

class lstset(Base.Command):
    args = 'arguments:dict'
    def invoke(self, tex):
        Base.Command.invoke(self, tex)
        if 'language' in list(self.attributes['arguments'].keys()):
            self.ownerDocument.context.current_language = \
                self.attributes['arguments']['language']

class lstlisting(Base.verbatim):
    args = '[ arguments:dict ]'
    counter = 'listings'

    def invoke(self, tex):
        if self.macroMode == Base.Environment.MODE_END:
            return
        s = ''.join(Base.verbatim.invoke(self, tex)[1:]).replace('\r','').split('\n')
        _format(self, s)

class lstinline(Base.verb):
    args = '[ arguments:dict ]'

    def invoke(self, tex):
        _format(self, ''.join(Base.verb.invoke(self, tex)[2:-1]))

class lstinputlisting(Base.Command):
    args = '[ arguments:dict ] file:str'
    counter = 'listings'

    def invoke(self, tex):
        Base.Command.invoke(self, tex)
        if 'file' not in list(self.attributes.keys()) or not self.attributes['file']:
            raise ValueError('Malformed \\lstinputlisting macro.')
        encoding = self.config['files']['input-encoding']
        _format(self, open(self.attributes['file'],  encoding=encoding))

def _format(self, file):
    if self.attributes['arguments'] is None:
        self.attributes['arguments'] = {}

    linenos = False
    if ('numbers' in list(self.attributes['arguments'].keys())
        or 'numbers' in list(PackageOptions.keys())):
        linenos = 'inline'

    # If this listing includes a label, inform plasTeX.
    if 'label' in list(self.attributes['arguments'].keys()):
        if hasattr(self.attributes['arguments']['label'], 'textContent'):
            self.ownerDocument.context.label(
                self.attributes['arguments']['label'].textContent)
        else:
            self.ownerDocument.context.label(
                self.attributes['arguments']['label'])

    # Check the textual LaTeX arguments and convert them to Python
    # attributes.
    if 'firstline' in list(self.attributes['arguments'].keys()):
        first_line_number = int(self.attributes['arguments']['firstline'])
    else:
        first_line_number = 0

    if 'lastline' in list(self.attributes['arguments'].keys()):
        last_line_number = int(self.attributes['arguments']['lastline'])
    else:
        last_line_number = sys.maxsize

    # Read the file, all the while respecting the "firstline" and
    # "lastline" arguments given in the document.
    self.plain_listing = ''
    for current_line_number, line in enumerate(file):
        current_line_number += 1
        if (current_line_number >= first_line_number) and \
           (current_line_number <= last_line_number):
            # Remove single-line "listings" comments. Only
            # comments started by "/*@" and ended by "@*/" are
            # supported.
            line = re.sub('/\*@[^@]*@\*/', '', line)

            # Add the just-read line to the listing.
            if hasattr(file, 'read'):
                self.plain_listing += line
            else:
                self.plain_listing += '\n' + line


    # Create a syntax highlighted XHTML version of the file using Pygments
    if pygments is not None:
        from pygments import lexers, formatters
        try:
            lexer = lexers.get_lexer_by_name(self.ownerDocument.context.current_language.lower())
        except Exception as msg:
            lexer = lexers.TextLexer()
        self.xhtml_listing = pygments.highlight(self.plain_listing, lexer, formatters.HtmlFormatter(linenos=linenos))



#!/usr/bin/env python

"""
C.3.3 Footnotes (p172)

"""

from plasTeX import Command, Environment, DimenCommand
from plasTeX.Logging import getLogger


class footnote(Command):
    args = '[ num:int ] self'
    mark = None

    def invoke(self, tex):
        # Add the footnote to the document
        output = Command.invoke(self, tex)
        userdata = self.ownerDocument.userdata
        if 'footnotes' not in userdata:
            userdata['footnotes'] = []
        userdata['footnotes'].append(self)
        self.mark = self
        return output

class footnotemark(Command):
    args = '[ num:int ]'
    mark = None

    def invoke(self, tex):
        # Add the footnotemarks to the document
        output = Command.invoke(self, tex)
        userdata = self.ownerDocument.userdata
        if 'footnotemarks' not in userdata:
            userdata['footnotemarks'] = []
        userdata['footnotemarks'].append(self)
        self.mark = self
        return output

class footnotetext(footnote):
    args = '[ num:int ] self'
    mark = None
    
    def invoke(self, tex):
        output = footnote.invoke(self, tex)
        self.mark = self.ownerDocument.userdata.get('footnotemarks',[None]).pop(0)
        return output

#
# Style Parameters
#

class footnotesep(DimenCommand):
    value = DimenCommand.new(0)

class footnoterule(Command):
    pass



#!/usr/bin/env python

from plasTeX import ismacro, macroName
from plasTeX.DOM import Node
from plasTeX.Logging import getLogger
from plasTeX.Tokenizer import Tokenizer, Token, DEFAULT_CATEGORIES, VERBATIM_CATEGORIES
import os
import configparser
import re
import time
import plasTeX

# Only export the Context singleton
__all__ = ['Context']

# Set up loggers
log = getLogger()
status = getLogger('status')
stacklog = getLogger('context.stack')
macrolog = getLogger('context.macros')


class ContextItem(dict):
    """
    Localized macro/category code stack element

    """

    def __init__(self, data=None):
        dict.__init__(self, data or {})
        self.categories = None
        self.obj = None
        self.parent = None
        self.owner = None

    @property
    def name(self):
        if self.obj is not None:
            return self.obj.nodeName
        return '{}'

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            if self.parent is not None and self.parent is not self:
                return self.parent[key]
            raise

    def get(self, key, default=None):
        try: return self[key]
        except KeyError: return default

    def has_key(self, key):
        if dict.__contains__(self, key):
            return True
        if self.parent is not None:
            return key in list(self.parent.keys())

    __contains__ = has_key

    def keys(self):
        keys = {}
        for key in list(dict.keys(self)):
            keys[key] = 0
        if self.parent is not None:
            for key in list(self.parent.keys()):
                keys[key] = 0
        return list(keys.keys())

    def __str__(self):
        if self.parent is not None:
             return '%s -> %s' % (self.parent, self.name)
        return self.name


class Counters(dict):
    def __getitem__(self, name):
        try:
            c = dict.__getitem__(self, name)
        except KeyError:
#           log.warning('No counter "%s" exists.  Creating one.' % name)
            c = self[name] = plasTeX.Counter(self.context, name)
        return c


class LanguageParser(object):
    """ Parser for language commands """

    def __init__(self, output=None):
        self.data = output if output is not None else {}
        self.language = None
        self.term = None

    def parse(self, files, encoding='UTF-8'):
        from xml.parsers import expat
        if isinstance(files, str):
            files = [files]
        for fname in files:
            if not os.path.isfile(fname):
                continue

            self.parser = expat.ParserCreate(encoding)
            self.parser.StartElementHandler = self.startElement
            self.parser.EndElementHandler = self.endElement
            self.parser.CharacterDataHandler = self.charData
            with open(fname, encoding=encoding) as f:
                self.parser.Parse(f.read())
        self.mergeLanguages()
        return self.data

    def mergeLanguages(self):
        # Merge language keys from the major language section, into
        # the minor language section
        for key, value in list(self.data.items()):
            if '-' in key:
                major, minor = key.split('-', 1)
                if major in list(self.data.keys()):
                    majordict = self.data[major]
                    for mkey, mvalue in list(majordict.items()):
                        if mkey not in value:
                            value[mkey] = mvalue

    def startElement(self, name, attrs):
        if name == 'terms':
            self.term = None
            if self.data.get(attrs['lang']):
                self.language = self.data[attrs['lang']]
            else:
                self.language = self.data[attrs['lang']] = {}
            if 'babel' in attrs:
                self.data[attrs['babel']] = self.language
        elif name == 'term':
            self.term = attrs['name']
            self.language[self.term] = ''

    def endElement(self, name):
        if name == 'term':
            self.term = None

    def charData(self, data):
        if self.term:
            self.language[self.term] += data


class Context(object):
    """
    Object to handle macro contexts within a TeX document

    This class keeps track of macros (both global and local), labels,
    context groupings, category codes, etc.  The TeX parser uses this
    class to hold any and all information about the document currently
    being processed.  This class also contains methods to generate
    new counters, ifs, dimensions, and other commands and macros.

    """

    def globals(self):
        return self.contexts[0]

    def __init__(self, load=False):
        # Stack of ContextItems
        self.contexts = []

        # Object that the current label points to
        self.currentlabel = None

        # Labeled objects
        self.labels = {}
        self.persistentLabels = {}

        # Unresolved refs
        self.refs = {}

        # LaTeX counters
        self.counters = Counters()
        self.counters.context = self

        # Tokens aliased by \let
        self.lets = {}

        # Imported packages and their options
        self.packages = {}

        # Output files
        self.writes = {}

        # Depth of the context stack
        self.depth = 0

        # Holds the current environment name stack
        self._currenvir = []

        # Holds the terms for various languages
        self.languages = {}
        self.terms = {}
        self.currentLanguage = ''

        # Create a global namespace
        self.push()

        # Meta information
        self.meta = {'lines': 0}

        self.warnOnUnrecognized = True

        if load:
            self.loadBaseMacros()

    def currenvir():
        def fget(self):
            if self._currenvir:
                return self._currenvir[-1]
            return
        def fset(self, value):
            if value is None:
                self._currenvir.pop()
            else:
                self._currenvir.append(value)
        def fdel(self):
            self._currenvir.pop()
        return locals()
    currenvir = property(**currenvir())

    def persist(self, filename, rtype='none'):
        """
        Persist cross-document information for labeled nodes

        Required Arguments:
        filename -- the name of the file with the shelved data

        Keyword Arguments:
        rtype -- the key in the shelved data to look under.  This is generally
            the name of the renderer used since the information for each
            renderer may be different.

        """
        import pickle
        if os.path.exists(filename):
            try:
                d = pickle.load(open(filename, 'rb'))
                if rtype not in list(d.keys()):
                    d[rtype] = {}
            except:
                os.remove(filename)
                d = {rtype:{}}
        else:
            d = {rtype:{}}
        data = d[rtype]
        for key, value in list(self.persistentLabels.items()):
            data[key] = value.persist()
        try:
            pickle.dump(d, open(filename, 'wb'))
        except Exception as msg:
            log.warning('Could not save auxiliary information. (%s)' % msg)

    def restore(self, filename, rtype='none'):
        """
        Restore cross-document information for labeled nodes

        Required Arguments:
        filename -- the name of the file with the shelved data

        Keyword Arguments:
        rtype -- the key in the shelved data to look under.  This is generally
            the name of the renderer used since the information for each
            renderer may be different.

        """
        import pickle
        if not os.path.exists(filename):
            return
        try:
            d = pickle.load(open(filename, 'rb'))
            try: data = d[rtype]
            except KeyError: return
            wou = self.warnOnUnrecognized
            self.warnOnUnrecognized = False
            for key, value in list(data.items()):
                n = self[value.get('macroName', 'Macro')]()
                n.restore(value)
                self.labels[key] = n
            self.warnOnUnrecognized = wou
        except Exception as msg:
            log.warning('Could not load auxiliary information. (%s)' % msg)

    @property
    def isMathMode(self):
        """ Are we in math mode or not? """
        for i in range(len(self.contexts) - 1, -1, -1):
            obj = self.contexts[i].obj
            if obj is not None and obj.mathMode is not None:
                return obj.mathMode
        return False

    def loadBaseMacros(self):
        """ Import all builtin macros """
        from plasTeX import Base
        self.importMacros(vars(Base))

    def loadLanguage(self, lang, document):
        """
        Load a localized version of macros for a particular language

        Required Arguments:
        lang -- the name of the language file to load

        """

        if not self.languages:
            files = document.config['document']['lang-terms'].split(os.pathsep)
            files.append(os.path.join(os.path.dirname(__file__), 'i18n.xml'))

            LanguageParser(self.languages).parse(reversed(files))

        if lang in list(self.languages.keys()):
            self.currentLanguage = lang
            self.newcommand('languagename', definition=lang)
            self.terms = self.languages[lang]
            for key, value in list(self.languages[lang].items()):
                if key == 'today':
                    self.newcommand(key, definition=self._strftime(value))
                else:
                    self.newcommand('%sname' % key, definition=value)
        else:
            log.warning('Could not load language "%s", american will be used instead' % lang)

    def _strftime(self, fmt):
        if '%f' in fmt or '%e' in fmt:
            day = time.strftime('%d')
            suffix = 'th'
            if day.endswith('1'):
                suffix = 'st'
            elif day.endswith('2'):
                suffix = 'nd'
            elif day.endswith('3'):
                suffix = 'rd'
            day = str(int(day))
            return time.strftime(fmt.replace('%f', day + suffix).replace('%e', day))
        return time.strftime(fmt)

    def loadINIPackage(self, inifile):
        """
        Load INI file containing macro definitions

        Arguments:
        inifile -- filename of INI formatted file

        """
        ini = configparser.RawConfigParser()
        if not isinstance(inifile, (list, tuple)):
            inifile = [inifile]
        for f in inifile:
            ini.read(f)
            macros = {}
            for section in ini.sections():
                try: baseclass = self[section]
                except KeyError:
                    log.warning('Could not find macro %s' % section)
                    continue
                for name in ini.options(section):
                    value = ini.get(section, name)
                    m = re.match(r'^str\(\s*(?:(\'|\")(?P<string>.+)(?:\1)|(?P<number>\d+))\s*\)$', value)
                    if m:
                        data = m.groupdict()
                        if data['number'] is not None:
                            value = chr(int(data['number']))
                        else:
                            value = data['string']
                        macros[name] = type(name, (baseclass,),
                                                    {'str': value})
                        continue
                    macros[name] = type(name, (baseclass,),
                                                {'args': value})
            self.importMacros(macros)

    def loadPackage(self, tex, file, options=None):
        """
        Load a Python or LaTeX package

        A Python version of the package is searched for first,
        if one cannot be found then a LaTeX version of the package
        is searched for.

        Required Arguments:
        tex -- the instance of the TeX engine to use for parsing
            the LaTeX file, if needed.
        file -- the name of the file to load

        Keyword Arguments:
        options -- the options given on the macro to pass to the package

        Returns:
        boolean indicating whether or not the package loaded successfully

        """
        options = options or {}
        module = os.path.splitext(file)[0]
        # See if it has already been loaded
        if module in list(self.packages.keys()):
            return True

        packagesini = os.path.join(os.path.dirname(plasTeX.Packages.__file__),
                                   os.path.basename(module) + '.ini')

        try:
            # Try to import a Python package by that name
            m = __import__(module, globals(), locals())
            status.info(' ( %s ' % m.__file__)
            if hasattr(m, 'ProcessOptions'):
                m.ProcessOptions(options, tex.ownerDocument)

            self.importMacros(vars(m))
            moduleini = os.path.splitext(m.__file__)[0] + '.ini'
            self.loadINIPackage([packagesini, moduleini])
            self.packages[module] = options
            status.info(' ) ')
            return True

        except ImportError as msg:
            log.warning('No Python version of %s was found' % file)
            # No Python module
            if 'No module' in str(msg) and module in str(msg):
                pass
                # Failed to load Python package
            # Error while importing
            else:
                raise

        result = tex.loadPackage(file, options)
        try:
            moduleini = os.path.join(os.path.dirname(tex.kpsewhich(file)),
                                     os.path.basename(module) + '.ini')
            self.loadINIPackage([packagesini, moduleini])
        except OSError: pass
        return result


    def label(self, label, node=None):
        """
        Set a label to the current labelable object

        Required Arguments:
        label -- string that contains the label

        Keyword Arguments:
        node -- a node to apply the label to rather than the currently
            labelable object

        See Also:
        self.ref()

        """
        label = label.strip()
        if not label:
            return

        if node is None:
            node = self.currentlabel

        if node is not None:
            self.persistentLabels[label] = self.labels[label] = node
            node.id = label

        # print label, ''.join(self.currentlabel.ref[:])

        # Resolve any outstanding references to this object
        if label in list(self.refs.keys()) and label in (self.labels.keys()):
            for obj in self.refs[label]:
                for key, value in list(obj.idref.items()):
                    if value.id != label:
                        continue
                    obj.idref[key] = self.labels[label]
            del self.refs[label]

    def ref(self, obj, name, label):
        """
        Set up a ref for resolution

        Required Arguments:
        obj -- object to put the referenced object onto
        name -- name of key in idref dictionary where object is stored
        label -- label to resolve

        See Also:
        self.label()

        """
        label = label.strip()
        if not label:
            return

        # Resolve ref if label already exists
        if label in list(self.labels.keys()):
            obj.idref[name] = self.labels[label]
            return

        # If the label doesn't exist, store away the object for later
        if label not in list(self.refs.keys()):
            self.refs[label] = []
        self.refs[label].append(obj)

        # Make a fake node with the ID on it for now
        node = self['Macro']()
        node.id = label
        obj.idref[name] = node

    def __getitem__(self, key):
        """
        Look through the stack of macros and return the requested one

        Required Arguments:
        key -- name of macro

        Returns: instance of requested macro

        """
        try: return self.top[key]
        except KeyError: pass

        # Didn't find it, so generate a new class
        if self.warnOnUnrecognized and not self.isMathMode:
            log.warning('unrecognized command/environment: %s', key)
        self[key] = newclass = type(key, (plasTeX.UnrecognizedMacro,), {})
        return newclass

    def push(self, context=None):
        """
        Add a new context to the stack

        This adds a new context grouping to the stack.  A context
        grouping includes both a set of localized macros and localized
        category codes.

        Keyword Arguments:
        context -- macro instance to use as the basis for a context
            grouping.  The local macros and category codes of this
            instance are used.  If this argument isn't supplied,
            an empty context is created.

        """
        if not self.contexts:
            context = ContextItem()
            context.categories = DEFAULT_CATEGORIES[:]
            self.contexts.append(context)

        else:
            name = '{}'
            if context is not None:
                name = context.nodeName
                # If we hit a document element, make sure that we start
                # at the global context.
                if context.level == context.DOCUMENT_LEVEL:
                    while len(self.contexts) > 1:
                        self.contexts.pop()
            stacklog.debug('pushing %s onto %s', name, self.top)
            self.contexts.append(self.createContext(context))

        self.mapMethods()

    append = push

    def __contains__(self, key):
        return key in self.top


    def mapMethods(self):
        # Getter methods use the most local context
        self.top = top = self.contexts[-1]
        self.__getitem__ = top.__getitem__
        self.__contains__ = top.__contains__
        self.get = top.get
        self.keys = top.keys
        self.has_key = top.has_key
        self.categories = top.categories

        # Setter methods always use the global namespace
        self.update = top.update
        self.__setitem__ = self.contexts[0].__setitem__

        # Set up inheritance attributes
        self.top.owner = self
        if len(self.contexts) > 1:
            self.top.parent = self.contexts[-2]

        self.depth = len(self.contexts)

    def createContext(self, obj=None):
        """
        Create the pieces of a new context (i.e. macros and category codes)

        Keyword Arguments:
        obj -- macro instance to use as the basis for a context
            grouping.  The local macros and category codes of this
            instance are used.  If this argument isn't supplied,
            an empty context is created.

        Returns: ContextItem instance

        """
        newcontext = ContextItem()
        newcontext.categories = self.categories
        newcontext.obj = obj

        if obj is not None:

            # Get the local category codes and macros
#           if obj.categories is not None:
#               newcontext.categories = obj.categories

            newcontext.update(obj.locals())

        return newcontext

    def importMacros(self, context):
        """
        Import macros from given context into the global namespace

        Required Arguments:
        context -- dictionary of macros to import

        """
        for value in list(context.values()):
            if ismacro(value):
                self[macroName(value)] = value
#           elif isinstance(value, Context):
#               self.importMacros(value)

    def pop(self, obj=None):
        """
        Remove a context from the stack

        Keyword Arguments:
        index -- index of context item to remove

        Returns: ContextItem instance removed from stack

        """
#       print 'POP', type(obj).__name__, self.contexts[-1]
        if obj is None:
            # Pop until we hit a None in the context
            while len(self.contexts) > 1:
                if self.contexts[-1].obj is None:
                    self.contexts.pop()
                    break
                self.contexts.pop()
        else:
            while len(self.contexts) > 1:
                o = self.contexts[-1].obj
                # If None, keep going
                if o is None:
                    pass
                # Found context pushed by ourself
                elif o is obj:
                    self.contexts.pop()
                    break
                # Don't pop parent node
                elif o is obj.parentNode:
                    break
                # Found the \begin to our \end
                elif type(obj) == type(o) and obj.macroMode == obj.MODE_END:
                    self.contexts.pop()
                    break
                # Found the \foo to our \endfoo
                elif obj.nodeName == ('end%s' % o.nodeName):
                    self.contexts.pop()
                    break
                self.contexts.pop()

        self.mapMethods()

    def addGlobal(self, key, value):
        """
        Add a macro to the global context

        Required Arguments:
        key -- name of macro to add
        value -- item to add to the global namespace.  If the item
            is a macro instance, it is simply added to the namespace.
            If it is a string, it is converted into a string Command
            instance before being added.

        """
        if isinstance(value, str):
            newvalue = plasTeX.Command()
            newvalue.str = value
            value = newvalue

        elif not ismacro(value):
            raise ValueError('"%s" does not implement the macro interface' % key)

        self.contexts[0][macroName(value)] = value

    __setitem__ = addGlobal

    def addLocal(self, key, value):
        """
        Add a macro to the local context

        Required Arguments:
        key -- name of macro to add
        value -- item to add to the global namespace.  If the item
            is a macro instance, it is simply added to the namespace.
            If it is a string, it is converted into a string Command
            instance before being added.

        """
        if isinstance(value, str):
            newvalue = plasTeX.Command()
            newvalue.str = value
            value = newvalue

        elif not ismacro(value):
            raise ValueError('"%s" does not implement the macro interface' % key)

        self.contexts[-1][macroName(value)] = value

    def whichCode(self, char):
        """
        Return the character code that `char` belongs to

        Required Arguments:
        char -- character to determine the code of

        Returns: integer category code of the given character

        """
        c = self.categories
        if char in c[Token.CC_LETTER]:
            return Token.CC_LETTER
        if char in c[Token.CC_SPACE]:
            return Token.CC_SPACE
        if char in c[Token.CC_EOL]:
            return Token.CC_EOL
        if char in c[Token.CC_BGROUP]:
            return Token.CC_BGROUP
        if char in c[Token.CC_EGROUP]:
            return Token.CC_EGROUP
        if char in c[Token.CC_ESCAPE]:
            return Token.CC_ESCAPE
        if char in c[Token.CC_SUPER]:
            return Token.CC_SUPER
        if char in c[Token.CC_SUB]:
            return Token.CC_SUB
        if char in c[Token.CC_MATHSHIFT]:
            return Token.CC_MATHSHIFT
        if char in c[Token.CC_ALIGNMENT]:
            return Token.CC_ALIGNMENT
        if char in c[Token.CC_COMMENT]:
            return Token.CC_COMMENT
        if char in c[Token.CC_ACTIVE]:
            return Token.CC_ACTIVE
        if char in c[Token.CC_PARAMETER]:
            return Token.CC_PARAMETER
        if char in c[Token.CC_IGNORED]:
            return Token.CC_IGNORED
        if char in c[Token.CC_INVALID]:
            return Token.CC_INVALID
        return Token.CC_OTHER

    def catcode(self, char, code):
        """
        Set the category code for a particular character

        Required arguments:
        char -- the character to set the code of
        code -- the category code number to set `char` to

        """
        c = self.contexts[-1].categories = self.categories = self.categories[:]
        for i in range(0, 16):
            c[i] = c[i].replace(char, '')
        # Don't insert if it's code 12.
        if code != 12:
            c[code] += char

    def setVerbatimCatcodes(self):
        """
        Set the category codes up for parsing verbatims

        This method turns the category codes for all characters to CC_OTHER

        """
        self.contexts[-1].categories = self.categories = VERBATIM_CATEGORIES[:]

    def newcounter(self, name, resetby=None, initial=0, format=None):
        """
        Create a new counter

        This method corresponds to LaTeX's \\newcounter command

        Required Arguments:
        name -- name of the counter.  The generate counter class will
            use this name.  Also, a new macro called 'the<name>' will
            also be generated for the counter format.

        Keyword Arguments:
        resetby -- the name of the counter that this counter is reset by
        initial -- initial value for the counter

        """
        # Counter already exists
        if name in list(self.counters.keys()):
            macrolog.debug('counter %s already defined', name)
            return
        self.counters[name] = plasTeX.Counter(self, name, resetby, initial)

        if format is None:
            format = '${%s}' % name
        newclass = type('the' + name, (plasTeX.TheCounter,),
                               {'format': format})
        self.addGlobal('the' + name, newclass)

    def newwrite(self, name, file):
        """
        Create a new output file

        Required Arguments:
        name -- the key name for the file
        file -- the file name to open

        """
        self.writes[name] = open(file, 'w')

    def newcount(self, name, initial=0):
        """
        Create a new count (like \\newcount)

        Required Arguments:
        name -- name of count to create

        Keyword Arguments:
        initial -- value to initialize to

        """
        # Generate a new count class
        macrolog.debug('creating count %s', name)
        newclass = type(name, (plasTeX.CountCommand,),
                               {'value': plasTeX.count(initial)})
        self.addGlobal(name, newclass)

    def newdimen(self, name, initial=0):
        """
        Create a new dimen (like \\newdimen)

        Required Arguments:
        name -- name of dimen to create

        Keyword Arguments:
        initial -- value to initialize to

        """
        # Generate a new dimen class
        macrolog.debug('creating dimen %s', name)
        newclass = type(name, (plasTeX.DimenCommand,),
                                {'value': plasTeX.dimen(initial)})
        self.addGlobal(name, newclass)

    def newskip(self, name, initial=0):
        """
        Create a new glue (like \\newskip)

        Required Arguments:
        name -- name of glue to create

        Keyword Arguments:
        initial -- value to initialize to

        """
        # Generate a new glue class
        macrolog.debug('creating dimen %s', name)
        newclass = type(name, (plasTeX.GlueCommand,),
                                {'value': plasTeX.glue(initial)})
        self.addGlobal(name, newclass)

    def newmuskip(self, name, initial=0):
        """
        Create a new muglue (like \\newmuskip)

        Required Arguments:
        name -- name of muglue to create

        Keyword Arguments:
        initial -- value to initialize to

        """
        # Generate a new muglue class
        macrolog.debug('creating muskip %s', name)
        newclass = type(name, (plasTeX.MuGlueCommand,),
                                {'value': plasTeX.muglue(initial)})
        self.addGlobal(name, newclass)

    def newif(self, name, initial=False):
        """
        Create a new \\if (and accompanying) commands

        This method corresponds to TeX's \\newif command.

        Required Arguments:
        name -- name of the 'if' command.  This name should always
            start with the letters 'if'.

        Keyword Arguments:
        initial -- initial value of the 'if' command

        """

        # \if already exists
        if name in list(self.keys()):
            macrolog.debug('if %s already defined', name)
            return

        # Generate new 'if' class
        macrolog.debug('creating if %s', name)
        ifclass = type(name, (plasTeX.NewIf,), {'state':initial})
        self.addGlobal(name, ifclass)

        # Create \iftrue macro
        truename = name[2:] + 'true'
        newclass = type(truename, (plasTeX.IfTrue,), {'ifclass':ifclass})
        self.addGlobal(truename, newclass)

        # Create \iffalse macro
        falsename = name[2:] + 'false'
        newclass = type(falsename, (plasTeX.IfFalse,), {'ifclass':ifclass})
        self.addGlobal(falsename, newclass)

    def newcommand(self, name, nargs=0, definition=None, opt=None):
        """
        Create a \\newcommand

        Required Arguments:
        name -- name of the macro to create
        nargs -- integer number of arguments that the macro has
        definition -- string containing the LaTeX definition
        opt -- string containing the LaTeX code to use in the
            optional argument

        Examples::
            c.newcommand('bold', 1, r'\\textbf{#1}')
            c.newcommand('foo', 2, r'{\\bf #1#2}', opt='myprefix')

        """
        # Macro already exists
        if name in list(self.keys()):
            if not issubclass(self[name], (plasTeX.NewCommand, plasTeX.Definition)):
                if not issubclass(self[name], plasTeX.TheCounter):
                    return
            macrolog.debug('redefining command "%s"', name)

        if nargs is None:
            nargs = 0
        assert isinstance(nargs, int), 'nargs must be an integer'

        if isinstance(definition, str):
            definition = [x for x in Tokenizer(definition, self)]

        if isinstance(opt, str):
            opt = [x for x in Tokenizer(opt, self)]

        macrolog.debug('creating newcommand %s', name)
        newclass = type(name, (plasTeX.NewCommand,),
                       {'nargs':nargs, 'opt':opt, 'definition':definition})

        self.addGlobal(name, newclass)

    def newenvironment(self, name, nargs=0, definition=None, opt=None):
        """
        Create a \\newenvironment

        Required Arguments:
        name -- name of the macro to create
        nargs -- integer number of arguments that the macro has
        definition -- two-element tuple containing the LaTeX definition.
            Each element should be a string.  The first element
            corresponds to the beginning of the environment, and the
            second element is the end of the environment.
        opt -- string containing the LaTeX code to use in the
            optional argument

        Examples::
            c.newenvironment('mylist', 0, (r'\\begin{itemize}', r'\\end{itemize}'))

        """
        # Macro already exists
        if name in list(self.keys()):
            if not issubclass(self[name], (plasTeX.NewCommand,
                                           plasTeX.Definition)):
                return
            macrolog.debug('redefining environment "%s"', name)

        if nargs is None:
            nargs = 0
        assert isinstance(nargs, int), 'nargs must be an integer'

        if definition is not None:
            assert isinstance(definition, (tuple, list)), \
                'definition must be a list or tuple'
            assert len(definition) == 2, 'definition must have 2 elements'

            if isinstance(definition[0], str):
                definition[0] = [x for x in Tokenizer(definition[0], self)]
            if isinstance(definition[1], str):
                definition[1] = [x for x in Tokenizer(definition[1], self)]

        if isinstance(opt, str):
            opt = [x for x in Tokenizer(opt, self)]

        macrolog.debug('creating newenvironment %s', name)

        # Begin portion
        newclass = type(name, (plasTeX.NewCommand,),
                       {'nargs':nargs, 'opt':opt, 'definition':definition[0]})
        self.addGlobal(name, newclass)

        # End portion
        newclass = type('end' + name, (plasTeX.NewCommand,),
                       {'nargs':0, 'opt':None, 'definition':definition[1]})
        self.addGlobal('end' + name, newclass)

    def newdef(self, name, args=None, definition=None, local=True):
        """
        Create a \def

        Required Arguments:
        name -- name of the macro to create
        args -- string containing the TeX argument profile
        definition -- string containing the LaTeX definition

        Keyword Arguments:
        local -- indicates whether this macro is local or global

        Examples::
            c.newdef('bold', '#1', '{\\bf #1}')
            c.newdef('put', '(#1,#2)#3', '\\dostuff{#1}{#2}{#3}')

        """
        # Macro already exists
#       if self.has_key(name):
#           if not issubclass(self[name], (plasTeX.NewCommand,
#                                          plasTeX.Definition)):
#               return
#           macrolog.debug('redefining definition "%s"', name)

        if isinstance(definition, str):
            definition = [x for x in Tokenizer(definition, self)]

        macrolog.debug('creating def %s', name)
        newclass = type(name, (plasTeX.Definition,),
                       {'args':args, 'definition':definition})

        if local:
            self.addLocal(name, newclass)
        else:
            self.addGlobal(name, newclass)

    def let(self, dest, source):
        """
        Create a \let

        Required Arguments:
        dest -- the command sequence to create
        source -- the token to set the command sequence equivalent to

        Examples::
            c.let('bgroup', BeginGroup('{'))

        """
        self.lets[dest] = source

    def chardef(self, name, num):
        """
        Create a \\chardef

        Required Arguments:
        name -- name of command to create
        num -- character number to use

        """
        # Generate a new chardef class
        macrolog.debug('creating chardef (8-bit) %s', name)
        newclass = type(name, (plasTeX.Command,),
                                {'str':chr(num)})
        self.addGlobal(name, newclass)

    def mathchardef(self, name, num):
        """
        Create a \\mathchardef

        Required Arguments:
        name -- name of command to create
        num -- character number to use

        """
        # Generate a new chardef class
        macrolog.debug('creating mathchardef (16-bit) %s', name)
        newclass = type(name, (plasTeX.Command,),
                                {'str':chr(num)})
        self.addGlobal(name, newclass)
